# -*- coding: utf-8 -*-
"""
Module to prepare the training datasets.
"""

from __future__ import print_function
import logging
import shutil
import math
from pathlib import Path
import pprint
from typing import List, Optional, Tuple, Union
import warnings
import urllib3

import geofileops as gfo
import pandas as pd
import geopandas as gpd
import numpy as np
import owslib
import owslib.wms
import owslib.util
from PIL import Image
import pygeos
import rasterio as rio
import rasterio.features as rio_features
import rasterio.profiles as rio_profiles
import shapely.geometry as sh_geom

from orthoseg.helpers.progress_helper import ProgressHelper
from orthoseg.util import ows_util
from orthoseg.util import vector_util

# -------------------------------------------------------------
# First define/init some general variables/constants
# -------------------------------------------------------------
# Get a logger...
logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)

# -------------------------------------------------------------
# The real work
# -------------------------------------------------------------


class LabelInfo:
    def __init__(
        self,
        locations_path: Path,
        polygons_path: Path,
        image_layer: str,
        locations_gdf: Optional[gpd.GeoDataFrame] = None,
        polygons_gdf: Optional[gpd.GeoDataFrame] = None,
    ):
        self.locations_path = locations_path
        self.polygons_path = polygons_path
        self.image_layer = image_layer
        self.locations_gdf = locations_gdf
        self.polygons_gdf = polygons_gdf

    def __repr__(self):
        repr = (
            f"LabelInfo with image_layer: {self.image_layer}, locations_path: "
            f"{self.locations_path}, polygons_path: {self.polygons_path}"
        )
        return repr


class ValidationError(ValueError):
    def __init__(self, message, errors):
        # Call the base class constructor with the parameters it needs
        super().__init__(message)

        # Now for your custom code...
        self.errors = errors

    def __repr__(self):
        repr = super().__repr__()
        if self.errors is not None and len(self.errors) > 0:
            repr += f"\n  -> Errors: {pprint.pformat(self.errors, indent=4, width=100)}"
        return repr

    def __str__(self):
        repr = super().__str__()
        if self.errors is not None and len(self.errors) > 0:
            repr += f"\n  -> Errors: {pprint.pformat(self.errors, indent=4, width=100)}"
        return repr


def prepare_traindatasets(
    label_infos: List[LabelInfo],
    classes: dict,
    image_layers: dict,
    training_dir: Path,
    labelname_column: str = "classname",
    image_pixel_x_size: float = 0.25,
    image_pixel_y_size: float = 0.25,
    image_pixel_width: int = 512,
    image_pixel_height: int = 512,
    ssl_verify: Union[bool, str] = True,
    force: bool = False,
) -> Tuple[Path, int]:
    """
    This function prepares training data for the vector labels provided.

    It will:
        * get orthophoto's from a WMS server
        * create the corresponding label mask for each orthophoto

    Returns a tuple with (output_dir, dataversion):
        output_dir: the dir where the traindataset was created/found
        dataversion: a version number for the dataset created/found

    Args
        label_infos (List[LabelInfo]): paths to the files with label polygons
            and locations to generate images for
        classes (dict): dict with the classes to detect as keys. The values
            are the following:
                - labelnames: list of labels to use for this class
                - weight:
                - burn_value:
        image_layers (dict):
        training_dir (Path):
        labelname_column (str): the column where the label names are stored in
            the polygon files. If the column name specified is not found, column
            "label_name" is used if it exists for backwards compatibility.
        ssl_verify (bool or str, optional): True to use the default
            certificate bundle as installed on your system. False disables
            certificate validation (NOT recommended!). If a path to a
            certificate bundle file (.pem) is passed, this will be used.
            In corporate networks using a proxy server this is often needed
            to evade CERTIFICATE_VERIFY_FAILED errors. Defaults to True.
    """
    # Init stuff
    # Interprete ssl_verify
    auth = None
    if ssl_verify is not None:
        # If it is a string, make sure it isn't actually a bool
        if isinstance(ssl_verify, str):
            if ssl_verify.lower() == "true":
                ssl_verify = True
            elif ssl_verify.lower() == "false":
                ssl_verify = False

        auth = owslib.util.Authentication(verify=ssl_verify)  # type: ignore
        if isinstance(ssl_verify, bool) and ssl_verify is False:
            urllib3.disable_warnings()
            logger.warn("SSL VERIFICATION IS TURNED OFF!!!")

    # Check if the first class is named "background"
    if len(classes) == 0:
        raise Exception("No classes specified")
    elif list(classes)[0].lower() != "background":
        classes_str = pprint.pformat(classes, sort_dicts=False, width=50)
        raise Exception(
            f"By convention, the first class must be called background!\n{classes_str}"
        )

    # Check if the latest version of training data is already ok
    # Determine the current data version based on existing output data dir(s),
    # If dir ends on _TMP_* ignore it, as it (probably) ended with an error.
    output_dirs = training_dir.glob("[0-9]*/")
    output_dirs = [
        output_dir for output_dir in output_dirs if "_TMP_" not in output_dir.name
    ]

    reuse_traindata = False
    if len(output_dirs) == 0:
        dataversion_new = 1
    else:
        # Get the output dir with the highest version (=first if sorted desc)
        output_dir_mostrecent = sorted(output_dirs, reverse=True)[0]
        dataversion_mostrecent = int(output_dir_mostrecent.name)

        # If none of the input files changed since previous run, reuse dataset
        for label_file in label_infos:
            reuse_traindata = True
            labellocations_output_mostrecent_path = (
                output_dir_mostrecent / label_file.locations_path.name
            )
            labeldata_output_mostrecent_path = (
                output_dir_mostrecent / label_file.polygons_path.name
            )
            if not (
                labellocations_output_mostrecent_path.exists()
                and labeldata_output_mostrecent_path.exists()
                and gfo.cmp(
                    label_file.locations_path, labellocations_output_mostrecent_path
                )
                and gfo.cmp(label_file.polygons_path, labeldata_output_mostrecent_path)
            ):
                reuse_traindata = False
                break
        if reuse_traindata:
            dataversion_new = dataversion_mostrecent
            logger.info(
                "Input label file(s) didn't change since last prepare_traindatasets, "
                f"so reuse version {dataversion_new}"
            )
        else:
            dataversion_new = dataversion_mostrecent + 1
            logger.info(
                "Input label file(s) changed since last prepare_traindatasets, "
                f"so create new training data version {dataversion_new}"
            )

    # Determine the output dir
    training_dataversion_dir = training_dir / f"{dataversion_new:02d}"

    # If the train data is already ok, just return
    if reuse_traindata is True:
        return (training_dataversion_dir, dataversion_new)

    # The input labels have changed: create new train version
    # -------------------------------------------------------
    # read, validate and prepare data
    labeldata = prepare_labeldata(
        label_infos=label_infos,
        classes=classes,
        labelname_column=labelname_column,
        image_pixel_x_size=image_pixel_x_size,
        image_pixel_y_size=image_pixel_y_size,
        image_pixel_width=image_pixel_width,
        image_pixel_height=image_pixel_height,
    )

    # Reading label data was succesfull, so prepare temp dir to put training dataset in.
    # A temp dir, so it can be removed/ignored if an error occurs later on.
    output_tmp_dir = create_tmp_dir(
        training_dir, f"{dataversion_new:02d}", remove_existing=True
    )

    # Copy the label input files to dest dir + to a backup dir.
    for label_file in label_infos:
        gfo.copy(label_file.locations_path, output_tmp_dir)
        gfo.copy(label_file.polygons_path, output_tmp_dir)
        backup_dir = label_file.locations_path.parent / f"backup_v{dataversion_new:02d}"
        shutil.rmtree(backup_dir, ignore_errors=True)
        backup_dir.mkdir(parents=True, exist_ok=True)
        gfo.copy(label_file.locations_path, backup_dir)
        gfo.copy(label_file.polygons_path, backup_dir)

    # Now create the images/masks for the new train version for the different traindata
    # types.
    traindata_types = ["train", "validation", "test"]
    nb_todo = 0
    for labellocations_gdf, _ in labeldata:
        labellocations_curr_gdf = labellocations_gdf[
            labellocations_gdf["traindata_type"].isin(traindata_types)
        ]
        nb_todo += len(labellocations_curr_gdf)
    progress = ProgressHelper(message="prepare training images", nb_steps_total=nb_todo)
    logger.info(f"Get images for {nb_todo} labels")

    for traindata_type in traindata_types:
        # Create output dirs...
        output_imagedatatype_dir = output_tmp_dir / traindata_type
        output_imagedata_image_dir = output_imagedatatype_dir / "image"
        output_imagedata_mask_dir = output_imagedatatype_dir / "mask"
        for dir in [
            output_imagedatatype_dir,
            output_imagedata_mask_dir,
            output_imagedata_image_dir,
        ]:
            if dir and not dir.exists():
                dir.mkdir(parents=True, exist_ok=True)

        for labellocations_gdf, labels_to_burn_gdf in labeldata:
            try:
                # Get the label locations for this traindata type
                labellocations_curr_gdf = labellocations_gdf[
                    labellocations_gdf["traindata_type"] == traindata_type
                ]

                # Loop trough all locations labels to get an image for each of them
                created_images_gdf = gpd.GeoDataFrame()
                created_images_gdf["geometry"] = None
                wms_imagelayer_layersources = {}
                for i, label_tuple in enumerate(labellocations_curr_gdf.itertuples()):

                    img_bbox = label_tuple.geometry
                    image_layer = getattr(label_tuple, "image_layer")

                    # If the wms to be used hasn't been initialised yet
                    if image_layer not in wms_imagelayer_layersources:
                        wms_imagelayer_layersources[image_layer] = []
                        for layersource in image_layers[image_layer]["layersources"]:
                            wms_service = owslib.wms.WebMapService(
                                url=layersource["wms_server_url"],
                                version=layersource["wms_version"],
                                auth=auth,
                            )
                            if layersource["wms_ignore_capabilities_url"]:
                                # If the wms url in capabilities should be ignored,
                                # overwrite with original url
                                nb = len(
                                    wms_service.getOperationByName("GetMap").methods
                                )
                                for method_id in range(nb):
                                    wms_service.getOperationByName("GetMap").methods[
                                        method_id
                                    ]["url"] = layersource["wms_server_url"]
                            wms_imagelayer_layersources[image_layer].append(
                                ows_util.LayerSource(
                                    wms_service=wms_service,
                                    layernames=layersource["layernames"],
                                    layerstyles=layersource["layerstyles"],
                                    bands=layersource["bands"],
                                    random_sleep=layersource["random_sleep"],
                                )
                            )

                    # Now really get the image
                    logger.debug(f"Get image for coordinates {img_bbox.bounds}")
                    assert labellocations_gdf.crs is not None
                    image_filepath = ows_util.getmap_to_file(
                        layersources=wms_imagelayer_layersources[image_layer],
                        output_dir=output_imagedata_image_dir,
                        crs=labellocations_gdf.crs,
                        bbox=img_bbox.bounds,  # type: ignore
                        size=(image_pixel_width, image_pixel_height),
                        image_format=ows_util.FORMAT_PNG,
                        # image_format_save=ows_util.FORMAT_TIFF,
                        image_pixels_ignore_border=image_layers[image_layer][
                            "image_pixels_ignore_border"
                        ],
                        transparent=False,
                        layername_in_filename=True,
                    )

                    # Create a mask corresponding with the image file
                    # image_filepath can be None if file exists, so check if not None...
                    if image_filepath is not None:
                        # Mask should not be in a lossy format!
                        mask_filepath = Path(
                            str(image_filepath)
                            .replace(
                                str(output_imagedata_image_dir),
                                str(output_imagedata_mask_dir),
                            )
                            .replace(".jpg", ".png")
                        )
                        nb_classes = len(classes)

                        # Only keep the labels that are meant for this image layer
                        labels_for_layer_gdf = (
                            labels_to_burn_gdf.loc[
                                labels_to_burn_gdf["image_layer"] == image_layer
                            ]
                        ).copy()
                        # assert to evade pyLance warning
                        if len(labels_for_layer_gdf) == 0:
                            logger.info(
                                f"No polygons to burn for image_layer {image_layer}!"
                            )
                        assert isinstance(labels_for_layer_gdf, gpd.GeoDataFrame)
                        _create_mask(
                            input_image_filepath=image_filepath,
                            output_mask_filepath=mask_filepath,
                            labels_to_burn_gdf=labels_for_layer_gdf,
                            nb_classes=nb_classes,
                            force=force,
                        )

                    # Log the progress and prediction speed
                    progress.step()

            except Exception as ex:
                raise ex

    # If everything went fine, rename output_tmp_dir to the final output_dir
    output_tmp_dir.rename(training_dataversion_dir)

    return (training_dataversion_dir, dataversion_new)


def prepare_labeldata(
    label_infos: List[LabelInfo],
    classes: dict,
    labelname_column: str,
    image_pixel_x_size: float,
    image_pixel_y_size: float,
    image_pixel_width: int,
    image_pixel_height: int,
) -> List[Tuple[gpd.GeoDataFrame, gpd.GeoDataFrame]]:
    """
    Prepare and validate the data in the labelinfos so it is ready to be uses to fetch
    train images and burn masks.

    Args:
        label_infos (List[LabelInfo]): the label files/data.
        classes (dict): dict with classes and their corresponding
            label class names + weights.
        labelname_column (str): the column name in the label polygon files where the
            label classname can be found. Defaults to "classname".

    Raises:
        ValidationError: if the data is somehow not valid. All issues are listed in the
            errors property.

    Returns:
        List[Tuple[gpd.GeoDataFrame, gpd.GeoDataFrame]]: returns a list of tuples with
            geodataframes with labellocations and labelpolygons to burn.
    """
    labeldata_result = []
    errors_found = []
    labellocations_found = False
    for label_info in label_infos:
        # Read label locations if needed
        if label_info.locations_gdf is not None:
            labellocations_gdf = label_info.locations_gdf
        else:
            logger.debug(f"Read label locations from {label_info.locations_path}")
            labellocations_gdf = gfo.read_file(label_info.locations_path)
            if labellocations_gdf is not None and len(labellocations_gdf) > 0:
                labellocations_gdf.loc[:, "path"] = str(label_info.locations_path)
                labellocations_gdf.loc[:, "image_layer"] = label_info.image_layer
                # Remark: geopandas 0.7.0 drops fid column internally!
                labellocations_gdf.loc[
                    :, "row_nb_orig"
                ] = labellocations_gdf.index  # type: ignore
            else:
                logger.warn(f"No label locations found in {label_info.locations_path}")
                continue

        if len(labellocations_gdf) > 0:
            labellocations_found = True

        # Read label polygons if needed
        if label_info.polygons_gdf is not None:
            labelpolygons_gdf = label_info.polygons_gdf
        else:
            logger.debug(f"Read label data from {label_info.polygons_path}")
            labelpolygons_gdf = gfo.read_file(label_info.polygons_path)
            if labelpolygons_gdf is not None and len(labelpolygons_gdf) > 0:
                labelpolygons_gdf.loc[:, "path"] = str(label_info.polygons_path)
                labelpolygons_gdf.loc[:, "image_layer"] = label_info.image_layer
            else:
                logger.warn(f"No label polygons found in {label_info.polygons_path}")

        assert labellocations_gdf is not None
        assert labelpolygons_gdf is not None

        if labellocations_gdf is None or len(labellocations_gdf) == 0:
            errors_found.append("No label locations found in labellocations_gdf")
            continue

        # Validate + process the location data
        # ------------------------------------
        if "traindata_type" not in labellocations_gdf.columns:
            errors_found.append(
                "Mandatory column traindata_type not found in "
                f"{label_info.locations_path}"
            )

        # Check + correct the geom of the location to make sure it matches the size of
        # the training images/masks to be generated...
        image_crs_width = math.fabs(
            image_pixel_width * image_pixel_x_size
        )  # tile width in units of crs => 500 m
        image_crs_height = math.fabs(
            image_pixel_height * image_pixel_y_size
        )  # tile height in units of crs => 500 m
        locations_none = []
        for location in labellocations_gdf.itertuples():
            if location.geometry is None or len(location.geometry.bounds) < 4:
                logger.warning(
                    f"No or empty geometry found in file {Path(location.path).name} "
                    f"for index {location.Index}, it will be ignored"
                )
                locations_none.append(location.Index)
                continue

            # Check if the traindata_type is valid
            if location.traindata_type not in ["train", "validation", "test", "todo"]:
                errors_found.append(
                    f"Invalid traindata_type in {Path(location.path).name}: "
                    f"{location.geometry.wkt}"
                )

            # Check if the geometry is valid
            is_valid_reason = pygeos.is_valid_reason(
                labellocations_gdf.geometry.array.data[location.Index]
            )
            if is_valid_reason != "Valid Geometry":
                errors_found.append(
                    f"Invalid geometry in {Path(location.path).name}: "
                    f"{is_valid_reason}"
                )
                continue

            geom_bounds = location.geometry.bounds
            xmin = geom_bounds[0] - (geom_bounds[0] % image_pixel_x_size)
            ymin = geom_bounds[1] - (geom_bounds[1] % image_pixel_y_size)
            xmax = xmin + image_crs_width
            ymax = ymin + image_crs_height
            location_geom_aligned = sh_geom.box(xmin, ymin, xmax, ymax)

            # Check if the realigned geom overlaps good enough with the original
            intersection = location_geom_aligned.intersection(
                labellocations_gdf.at[location.Index, "geometry"]
            )
            area_1row_1col = (
                image_pixel_x_size * image_crs_width
                + image_pixel_y_size * image_crs_height
            )
            if intersection.area < (location_geom_aligned.area - area_1row_1col):
                # Original geom was digitized too small
                errors_found.append(
                    f"Location geometry too small in {Path(location.path).name}: "
                    f"{location.geometry.wkt}"
                )
            elif location_geom_aligned.area > 1.1 * sh_geom.box(*geom_bounds).area:
                logger.warn(
                    "Location geometry larger than expected in file "
                    f"{Path(location.path).name}: {location.geometry.wkt}"
                )
            labellocations_gdf.at[location.Index, "geometry"] = location_geom_aligned

        # Remove locations with None or point/line geoms
        labellocations_gdf = labellocations_gdf[
            ~labellocations_gdf.index.isin(locations_none)
        ]  # type: ignore

        # Check if labellocations has a proper crs
        if labellocations_gdf.crs is None:
            errors_found.append(
                "No crs in labellocations, labellocation_gdf.crs: "
                f"{labellocations_gdf.crs}"
            )

        # Validate + process the polygons data
        # ------------------------------------
        if labelpolygons_gdf is None:
            errors_found.append("No labelpolygons in the training data!")
            continue

        # Create list with only the input polygons that need to be burned in the mask
        labels_to_burn_gdf = None
        if labelname_column not in labelpolygons_gdf.columns:
            # For backwards compatibility, also support old default column name
            labelname_column = "label_name"
        if labelname_column in labelpolygons_gdf.columns:
            # If there is a column labelname_column, use the burn values specified in
            # the configuration
            labels_to_burn_gdf = labelpolygons_gdf
            labels_to_burn_gdf.loc[:, "burn_value"] = None  # type: ignore
            for classname in classes:
                labels_to_burn_gdf.loc[
                    (
                        labels_to_burn_gdf[labelname_column].isin(
                            classes[classname]["labelnames"]
                        )
                    ),
                    "burn_value",
                ] = classes[classname]["burn_value"]

            # Check if there are invalid class names
            invalid_gdf = labels_to_burn_gdf.loc[
                labels_to_burn_gdf["burn_value"].isnull()
            ]
            for _, invalid_row in invalid_gdf.iterrows():
                errors_found.append(
                    f"Invalid classname in {Path(invalid_row['path']).name}: "
                    f"{invalid_row[labelname_column]}"
                )

            # Filter away rows that are going to burn 0, as this is useless...
            labels_to_burn_gdf = labels_to_burn_gdf.loc[
                labels_to_burn_gdf["burn_value"] != 0
            ].copy()

        elif len(classes) == 2:
            # There is no column with label names, but there are only 2 classes
            # (background + subject), so no problem...
            logger.info(
                f"Column ({labelname_column}) not found, so use all polygons in "
                f"{label_info.polygons_path.name}"
            )
            labels_to_burn_gdf = labelpolygons_gdf
            labels_to_burn_gdf.loc[:, "burn_value"] = 1  # type: ignore

        else:
            # There is no column with label names, but more than two classes, so stop.
            errors_found.append(
                f"Column {labelname_column} is mandatory in labeldata if multiple "
                f"classes specified: {classes}"
            )

        # Check if we ended up with label data to burn.
        if labels_to_burn_gdf is None:
            errors_found.append(
                "Not any labelpolygon retained to burn in the training data!"
            )
            continue

        # Filter away None and empty geometries... they cannot be burned
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", "GeoSeries.notna", UserWarning)
            labels_to_burn_gdf = labels_to_burn_gdf[
                ~labels_to_burn_gdf.geometry.is_empty
                & labels_to_burn_gdf.geometry.notna()
            ].copy()

        # Make sure all label polygons are valid
        labels_to_burn_gdf.loc[  # type: ignore
            :, "is_valid_reason"
        ] = vector_util.is_valid_reason(labels_to_burn_gdf.geometry)
        invalid_gdf = labels_to_burn_gdf.query("is_valid_reason != 'Valid Geometry'")
        for invalid in invalid_gdf.itertuples():
            errors_found.append(
                f"Invalid geometry in {Path(invalid.path).name}: "
                f"{invalid.is_valid_reason}"
            )

        assert isinstance(labels_to_burn_gdf, gpd.GeoDataFrame)
        labeldata_result.append((labellocations_gdf, labels_to_burn_gdf))

    # If errors found, raise
    if len(errors_found) > 0:
        raise ValidationError(
            f"Errors found in label data: {len(errors_found)}", errors_found
        )
    if not labellocations_found:
        errors_found.append("No labellocations found in the training data")
        raise ValidationError("No labellocations found", errors_found)

    return labeldata_result


def create_tmp_dir(
    parent_dir: Path, dir_name: str, remove_existing: bool = False
) -> Path:
    """
    Helper function to create a 'TMP' dir based on a directory name:
        parent_dir / <dir_name>_TMP_<sequence>

    Use: if you want to write data to a directory in a "transactional way",
    it is the safest to write to a temp dir first, and then rename it to the
    final name. This way, if a hard crash occurs while writing the data, it
    is clear that the directory wasn't ready. Additionally, in case of a hard
    crash, file locks can remain which makes it impossible to remove a
    directory for a while.

    Args:
        parent_dir (Path): The dir to create the temp dir in.
        dir_name (str): The name of the dir to base the temp dir on.
        remove_existing (bool, optional): If True, existing TMP directories
            will be removed if possible. Defaults to False.

    Raises:
        Exception: [description]

    Returns:
        Path: [description]
    """
    # The final dir should not exists yet!
    final_dir = parent_dir / dir_name
    if final_dir.exists():
        raise Exception(
            f"It is not supported to create a TMP dir for an existing dir: {final_dir}"
        )

    # Try to delete all existing TMP dir's
    if remove_existing is True:
        existing_tmp_dirs = parent_dir.glob(f"{dir_name}_TMP_*")
        for existing_tmp_dir in existing_tmp_dirs:
            try:
                shutil.rmtree(existing_tmp_dir)
            except Exception:
                tmp_dir = None

    # Create the TMP dir
    tmp_dir = None
    for i in range(100):
        tmp_dir = parent_dir / f"{dir_name}_TMP_{i:02d}"

        if tmp_dir.exists():
            # If it (still) exists try next sequence
            tmp_dir = None
            continue
        else:
            # If it doesn't exist, try to create it
            try:
                tmp_dir.mkdir(parents=True)
                break
            except Exception:
                # If it fails to create, try next sequence
                tmp_dir = None
                continue

    # If no output tmp dir could be found/created... stop...
    if tmp_dir is None:
        raise Exception(
            f"Error creating/replacing TMP dir for {dir_name} in {parent_dir}"
        )

    return tmp_dir


def _create_mask(
    input_image_filepath: Path,
    output_mask_filepath: Path,
    labels_to_burn_gdf: gpd.GeoDataFrame,
    nb_classes: int = 1,
    output_imagecopy_filepath: Optional[Path] = None,
    minimum_pct_labeled: float = 0.0,
    force: bool = False,
) -> Optional[bool]:

    # If file exists already and force is False... stop.
    if force is False and output_mask_filepath.exists():
        logger.debug(
            f"Output file exist, and force is False, return: {output_mask_filepath}"
        )
        return

    # Create a mask corresponding with the image file
    # First read the properties of the input image to copy them for the output
    logger.debug(f"Create mask to {output_mask_filepath}")
    with rio.open(input_image_filepath) as image_ds:
        image_input_profile = image_ds.profile
        image_transform_affine = image_ds.transform

    # Prepare the file profile for the mask depending on output type
    output_ext_lower = output_mask_filepath.suffix.lower()
    if output_ext_lower == ".tif":
        image_output_profile = rio_profiles.DefaultGTiffProfile(
            count=1, transform=image_transform_affine, crs=image_input_profile["crs"]
        )
    if output_ext_lower == ".png":
        image_output_profile = rio_profiles.Profile(driver="PNG", count=1)
    else:
        raise Exception(
            f"Unsupported mask suffix (should be lossless format!): {output_ext_lower}"
        )
    image_output_profile.update(
        width=image_input_profile["width"],
        height=image_input_profile["height"],
        dtype=rio.uint8,
    )

    # Burn the vectors in a mask
    burn_shapes = [
        (geom, value)
        for geom, value in zip(
            labels_to_burn_gdf.geometry, labels_to_burn_gdf.burn_value
        )
        if geom is not None and geom.is_empty is False
    ]
    if len(burn_shapes) > 0:
        try:
            mask_arr = rio_features.rasterize(
                shapes=burn_shapes,
                transform=image_transform_affine,
                dtype=rio.uint8,
                fill=0,
                out_shape=(
                    image_output_profile["width"],  # type: ignore
                    image_output_profile["height"],  # type: ignore
                ),
            )

        except Exception as ex:
            raise Exception(f"Error creating mask for {image_transform_affine}") from ex
    else:
        mask_arr = np.zeros(
            shape=(
                image_output_profile["width"],  # type: ignore
                image_output_profile["height"],  # type: ignore
            ),
            dtype=rio.uint8,
        )

    # Check if the mask meets the requirements to be written...
    if minimum_pct_labeled > 0:
        nb_pixels = np.size(mask_arr, 0) * np.size(mask_arr, 1)
        nb_pixels_data = nb_pixels - np.sum(
            mask_arr == 0
        )  # np.count_nonnan(image == NoData_value)
        logger.debug(
            f"nb_pixels: {nb_pixels}, nb_pixels_data: {nb_pixels_data}, "
            f"pct data: {nb_pixels_data / nb_pixels}"
        )

        if nb_pixels_data / nb_pixels < minimum_pct_labeled:
            return False

    # Write the labeled mask as .png (so without transform/crs info)
    im = Image.fromarray(mask_arr)
    im.save(output_mask_filepath)

    return True



# -*- coding: utf-8 -*-
"""
Tests for functionalities in orthoseg.lib.postprocess_predictions.
"""
import os
from pathlib import Path
import sys

import geofileops as gfo
import pandas as pd

# Make hdf5 version warning non-blocking
os.environ["HDF5_DISABLE_VERSION_CHECK"] = "1"

# Add path so the local orthoseg packages are found
sys.path.insert(0, str(Path(__file__).resolve().parent.parent))
from orthoseg.lib import postprocess_predictions as post_pred
from tests.test_helper import TestData


def test_read_prediction_file():
    # Read + polygonize raster prediction file
    pred_raster_path = (
        TestData.testdata_dir / "129568_185248_130592_186272_4096_4096_1_pred.tif"
    )
    pred_raster_gdf = post_pred.read_prediction_file(
        pred_raster_path, border_pixels_to_ignore=128
    )
    # gfo.to_file(pred_raster_gdf, get_testdata_dir() / f"{pred_raster_path.stem}.gpkg")

    # Read the comparison file, that contains the result of the polygonize
    pred_comparison_path = TestData.testdata_dir / f"{pred_raster_path.stem}.gpkg"
    pred_comparison_gdf = gfo.read_file(pred_comparison_path)

    # Now compare they are the same
    assert pred_raster_gdf is not None
    assert len(pred_raster_gdf) == len(pred_comparison_gdf)


def test_clean_vectordata(tmpdir):
    temp_dir = Path(tmpdir)

    # Clean data
    input1_path = (
        TestData.testdata_dir / "129568_184288_130592_185312_4096_4096_1_pred.gpkg"
    )
    input2_path = (
        TestData.testdata_dir / "129568_185248_130592_186272_4096_4096_1_pred.gpkg"
    )
    input1_gdf = gfo.read_file(input1_path)
    input2_gdf = gfo.read_file(input2_path)
    input_gdf = pd.concat([input1_gdf, input2_gdf])
    assert input1_gdf.crs == input_gdf.crs
    input_path = temp_dir / "vector_input.gpkg"
    gfo.to_file(input_gdf, input_path)
    output_path = temp_dir / input_path.name
    post_pred.postprocess_predictions(
        input_path=input_path,
        output_path=output_path,
        dissolve=True,
        dissolve_tiles_path=None,
        force=True,
    )

    # Read result and check
    geoms_simpl_filepath = (
        output_path.parent / f"{output_path.stem}_dissolve{output_path.suffix}"
    )
    result_gdf = gfo.read_file(geoms_simpl_filepath)

    assert len(result_gdf) == 616



# -*- coding: utf-8 -*-
"""
Module with specific helper functions to manage the logging of orthoseg.

TODO: maybe it is cleaner to replace most code here by a config dict?
"""

from email.message import EmailMessage
import logging
import smtplib
from typing import Optional

from orthoseg.helpers import config_helper as conf

# -------------------------------------------------------------
# First define/init general variables/constants
# -------------------------------------------------------------
# Get a logger...
logger = logging.getLogger(__name__)

# -------------------------------------------------------------
# The real work
# -------------------------------------------------------------


def sendmail(subject: str, body: Optional[str] = None, stop_on_error: bool = False):

    if conf is None:
        raise Exception("Config is not initialized")

    if not conf.email.getboolean("enabled", fallback=False):
        return
    mail_from = conf.email.get("from", None)
    mail_to = conf.email.get("to", None)
    mail_server = conf.email.get("server", None)
    mail_server_username = conf.email.get("username", None)
    mail_server_password = conf.email.get("password", None)

    # If one of the necessary parameters not provided, log subject
    if mail_from is None or mail_to is None or mail_server is None:
        logger.warning(
            f"Mail global_config not provided to send email with subject: {subject}"
        )
        return

    try:
        # Create message
        msg = EmailMessage()
        msg.add_header("from", mail_from)
        msg.add_header("to", mail_to)
        msg.add_header("subject", subject)
        if body is not None:
            msg.set_payload(body)
            msg.add_header("Content-Type", "text/html")

        # Send the email
        server = smtplib.SMTP(mail_server)
        if mail_server_username is not None and mail_server_password is not None:
            server.login(mail_server_username, mail_server_password)
        server.send_message(msg)
        server.quit()
    except Exception as ex:
        if stop_on_error is False:
            logger.exception("Error sending email")
        else:
            raise Exception("Error sending email") from ex



# -*- coding: utf-8 -*-
"""
Download the sample project.
"""

import argparse
from pathlib import Path
import sys

# orthoseg is higher in dir hierarchy, add root to sys.path
sys.path.insert(0, str(Path(__file__).resolve().parent.parent))
from orthoseg.util import gdrive_util
from orthoseg.util import git_downloader


def main():

    # Define supported arguments
    parser = argparse.ArgumentParser(add_help=False)

    help = (
        "The directory to download the orthoseg sample projects to. "
        "Use ~ for your home directory."
    )
    parser.add_argument("dest_dir", help=help)

    help = (
        "True to use the default certificate bundle as installed on your system. "
        "False disables certificate validation (NOT recommended!). In corporate "
        "networks using a proxy server it is often needed to specify a customized "
        "certificate bundle (.pem file) to avoid CERTIFICATE_VERIFY_FAILED errors. "
        "It is recommended to specify the path to a custum certificate bundle file "
        "using the REQUESTS_CA_BUNDLE environment variable, but it can also passed "
        "using this switch. Parameter defaults to True."
    )
    parser.add_argument("--ssl_verify", default=True, help=help)

    # Interprete arguments
    args = parser.parse_args()
    dest_dir = Path(args.dest_dir).expanduser() / "orthoseg"
    ssl_verify = args.ssl_verify

    dest_dir_full = dest_dir / "sample_projects"
    if dest_dir_full.exists():
        raise Exception(f"Destination directory already exists: {dest_dir_full}")

    # Download
    print(f"Start download of sample projects to {str(dest_dir_full)}")
    git_downloader.download(
        repo_url="https://github.com/orthoseg/orthoseg/tree/master/sample_projects",
        output_dir=dest_dir,
        ssl_verify=ssl_verify,
    )
    print("Download finished")
    print("Start download of footballfields pretrained neural net")
    footballfields_model_dir = dest_dir_full / "footballfields/models"
    model_hdf5_path = footballfields_model_dir / "footballfields_01_0.92512_242.hdf5"
    if model_hdf5_path.exists() is False:
        gdrive_util.download_file(
            "1XmAenCW6K_RVwqC6xbkapJ5ws-f7-QgH",
            model_hdf5_path,
            ssl_verify=ssl_verify,
        )
    model_hyperparams_path = (
        footballfields_model_dir / "footballfields_01_hyperparams.json"
    )
    if model_hyperparams_path.exists() is False:
        gdrive_util.download_file(
            "1umxcd4RkB81sem9PdIpLoWeiIW8ga1u7",
            model_hyperparams_path,
            ssl_verify=ssl_verify,
        )
    model_modeljson_path = footballfields_model_dir / "footballfields_01_model.json"
    if model_modeljson_path.exists() is False:
        gdrive_util.download_file(
            "16qe8thBTrO3dFfLMU1T22gWcfHVXt8zQ",
            model_modeljson_path,
            ssl_verify=ssl_verify,
        )
    print("Download finished")


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""
Tests for functionalities in orthoseg.train.
"""

from pathlib import Path
import sys

# Add path so the local orthoseg packages are found
sys.path.insert(0, str(Path(__file__).resolve().parent.parent))
import orthoseg
from tests.test_helper import TestData


def test_search_label_files():
    labeldata_template = (
        TestData.testdata_dir / "footballfields_{image_layer}_data.gpkg"
    )
    labellocation_template = (
        TestData.testdata_dir / "footballfields_{image_layer}_locations.gpkg"
    )
    result = orthoseg._search_label_files(labeldata_template, labellocation_template)

    assert len(result) == 2



# -*- coding: utf-8 -*-
"""
Module with functions for post-processing prediction masks towards polygons.
"""

import argparse
import logging
from pathlib import Path
import shlex
import sys
import traceback

import geofileops as gfo

# Because orthoseg isn't installed as package + it is higher in dir hierarchy, add root
sys.path.insert(0, str(Path(__file__).resolve().parent.parent))
from orthoseg.helpers import config_helper as conf
from orthoseg.helpers import email_helper
from orthoseg.lib import postprocess_predictions as postp
import orthoseg.model.model_helper as mh
from orthoseg.util import log_util

# -------------------------------------------------------------
# First define/init general variables/constants
# -------------------------------------------------------------
# Get a logger...
logger = logging.getLogger(__name__)

# -------------------------------------------------------------
# The real work
# -------------------------------------------------------------


def postprocess_argstr(argstr):
    args = shlex.split(argstr)
    postprocess_args(args)


def postprocess_args(args):

    # Interprete arguments
    parser = argparse.ArgumentParser(add_help=False)

    # Required arguments
    required = parser.add_argument_group("Required arguments")
    required.add_argument(
        "-c", "--config", type=str, required=True, help="The config file to use"
    )

    # Optional arguments
    optional = parser.add_argument_group("Optional arguments")
    # Add back help
    optional.add_argument(
        "-h",
        "--help",
        action="help",
        default=argparse.SUPPRESS,
        help="Show this help message and exit",
    )

    # Interprete arguments
    args = parser.parse_args(args)

    # Run!
    postprocess(config_path=Path(args.config))


def postprocess(config_path: Path):
    """
    Postprocess the output of a prediction for the config specified.

    Args:
        config_path (Path): Path to the config file.
    """

    # Init
    # Load the config and save in a bunch of global variables zo it
    # is accessible everywhere
    conf.read_orthoseg_config(config_path)

    # Init logging
    log_util.clean_log_dir(
        log_dir=conf.dirs.getpath("log_dir"),
        nb_logfiles_tokeep=conf.logging.getint("nb_logfiles_tokeep"),
    )
    global logger
    logger = log_util.main_log_init(conf.dirs.getpath("log_dir"), __name__)

    # Log start + send email
    message = f"Start postprocess for config {config_path.stem}"
    logger.info(message)
    logger.debug(f"Config used: \n{conf.pformat_config()}")
    email_helper.sendmail(message)

    try:

        # Create base filename of model to use
        # TODO: is force data version the most logical, or rather implement
        #       force weights file or ?
        traindata_id = None
        force_model_traindata_id = conf.train.getint("force_model_traindata_id")
        if force_model_traindata_id is not None and force_model_traindata_id > -1:
            traindata_id = force_model_traindata_id

        # Get the best model that already exists for this train dataset
        trainparams_id = conf.train.getint("trainparams_id")
        best_model = mh.get_best_model(
            model_dir=conf.dirs.getpath("model_dir"),
            segment_subject=conf.general["segment_subject"],
            traindata_id=traindata_id,
            trainparams_id=trainparams_id,
        )
        if best_model is None:
            raise Exception(f"No best model found in {conf.dirs.getpath('model_dir')}")

        # Input file  the "most recent" prediction result dir for this subject
        output_vector_dir = conf.dirs.getpath("output_vector_dir")
        output_vector_name = (
            f"{best_model['basefilename']}_{best_model['epoch']}_"
            f"{conf.predict['image_layer']}"
        )
        output_vector_path = output_vector_dir / f"{output_vector_name}.gpkg"

        # Prepare some parameters for the postprocessing
        nb_parallel = conf.general.getint("nb_parallel")

        dissolve = conf.postprocess.getboolean("dissolve")
        dissolve_tiles_path = conf.postprocess.getpath("dissolve_tiles_path")
        simplify_algorithm = conf.postprocess.get("simplify_algorithm")
        if simplify_algorithm is not None:
            simplify_algorithm = gfo.SimplifyAlgorithm[simplify_algorithm]
        simplify_tolerance = conf.postprocess.geteval("simplify_tolerance")
        simplify_lookahead = conf.postprocess.get("simplify_lookahead")
        if simplify_lookahead is not None:
            simplify_lookahead = int(simplify_lookahead)

        # Go!
        postp.postprocess_predictions(
            input_path=output_vector_path,
            output_path=output_vector_path,
            dissolve=dissolve,
            dissolve_tiles_path=dissolve_tiles_path,
            simplify_algorithm=simplify_algorithm,
            simplify_tolerance=simplify_tolerance,
            simplify_lookahead=simplify_lookahead,
            nb_parallel=nb_parallel,
            force=False,
        )

        # Log and send mail
        message = f"Completed postprocess for config {config_path.stem}"
        logger.info(message)
        email_helper.sendmail(message)
    except Exception as ex:
        message = f"ERROR while running postprocess for task {config_path.stem}"
        logger.exception(message)
        email_helper.sendmail(
            subject=message, body=f"Exception: {ex}\n\n {traceback.format_exc()}"
        )
        raise Exception(message) from ex


def main():
    try:
        postprocess_args(sys.argv[1:])
    except Exception as ex:
        logger.exception(f"Error: {ex}")
        raise


# If the script is ran directly...
if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""
Unet definition in Keras.

Based on code from:
    * unet: https://github.com/zhixuhao/unet
    * preload first layers with VGG16 weights:
      https://www.microsoft.com/developerblog/2018/07/05/satellite-images-segmentation-sustainable-farming/

TODO: review code and cleanup due to introduction of model factory

"""

import logging
from typing import Optional

import numpy as np
import tensorflow as tf
from tensorflow import keras as kr

# -------------------------------------------------------------
# First define/init some general variables/constants
# -------------------------------------------------------------
# Get a logger...
logger = logging.getLogger(__name__)

# -------------------------------------------------------------
# The real work
# -------------------------------------------------------------


def get_model(
    input_width=256,
    input_height=256,
    nb_channels=3,
    nb_classes=1,
    loss_mode="binary_crossentropy",
    learning_rate=0.0001,
    init_model_weights: bool = False,
    pretrained_weights_filepath: Optional[str] = None,
):

    inputs = kr.layers.Input((input_width, input_height, nb_channels))
    conv1 = kr.layers.Conv2D(
        64,
        3,
        activation="relu",
        padding="same",
        kernel_initializer="he_normal",
        name="block1_conv1",
    )(inputs)
    conv1 = kr.layers.Conv2D(
        64,
        3,
        activation="relu",
        padding="same",
        kernel_initializer="he_normal",
        name="block1_conv2",
    )(conv1)
    pool1 = kr.layers.MaxPooling2D(pool_size=(2, 2))(conv1)
    conv2 = kr.layers.Conv2D(
        128,
        3,
        activation="relu",
        padding="same",
        kernel_initializer="he_normal",
        name="block2_conv1",
    )(pool1)
    conv2 = kr.layers.Conv2D(
        128,
        3,
        activation="relu",
        padding="same",
        kernel_initializer="he_normal",
        name="block2_conv2",
    )(conv2)
    pool2 = kr.layers.MaxPooling2D(pool_size=(2, 2))(conv2)
    conv3 = kr.layers.Conv2D(
        256,
        3,
        activation="relu",
        padding="same",
        kernel_initializer="he_normal",
        name="block3_conv1",
    )(pool2)
    conv3 = kr.layers.Conv2D(
        256,
        3,
        activation="relu",
        padding="same",
        kernel_initializer="he_normal",
        name="block3_conv2",
    )(conv3)
    pool3 = kr.layers.MaxPooling2D(pool_size=(2, 2))(conv3)
    conv4 = kr.layers.Conv2D(
        512,
        3,
        activation="relu",
        padding="same",
        kernel_initializer="he_normal",
        name="block4_conv1",
    )(pool3)
    conv4 = kr.layers.Conv2D(
        512,
        3,
        activation="relu",
        padding="same",
        kernel_initializer="he_normal",
        name="block4_conv2",
    )(conv4)
    drop4 = kr.layers.Dropout(0.5)(conv4)
    pool4 = kr.layers.MaxPooling2D(pool_size=(2, 2))(drop4)

    conv5 = kr.layers.Conv2D(
        1024, 3, activation="relu", padding="same", kernel_initializer="he_normal"
    )(pool4)
    conv5 = kr.layers.Conv2D(
        1024, 3, activation="relu", padding="same", kernel_initializer="he_normal"
    )(conv5)
    drop5 = kr.layers.Dropout(0.5)(conv5)

    up6 = kr.layers.Conv2D(
        512, 2, activation="relu", padding="same", kernel_initializer="he_normal"
    )(kr.layers.UpSampling2D(size=(2, 2))(drop5))
    merge6 = kr.layers.Concatenate(axis=3)([drop4, up6])
    conv6 = kr.layers.Conv2D(
        512, 3, activation="relu", padding="same", kernel_initializer="he_normal"
    )(merge6)
    conv6 = kr.layers.Conv2D(
        512, 3, activation="relu", padding="same", kernel_initializer="he_normal"
    )(conv6)

    up7 = kr.layers.Conv2D(
        256, 2, activation="relu", padding="same", kernel_initializer="he_normal"
    )(kr.layers.UpSampling2D(size=(2, 2))(conv6))
    merge7 = kr.layers.Concatenate(axis=3)([conv3, up7])
    conv7 = kr.layers.Conv2D(
        256, 3, activation="relu", padding="same", kernel_initializer="he_normal"
    )(merge7)
    conv7 = kr.layers.Conv2D(
        256, 3, activation="relu", padding="same", kernel_initializer="he_normal"
    )(conv7)

    up8 = kr.layers.Conv2D(
        128, 2, activation="relu", padding="same", kernel_initializer="he_normal"
    )(kr.layers.UpSampling2D(size=(2, 2))(conv7))
    merge8 = kr.layers.Concatenate(axis=3)([conv2, up8])
    conv8 = kr.layers.Conv2D(
        128, 3, activation="relu", padding="same", kernel_initializer="he_normal"
    )(merge8)
    conv8 = kr.layers.Conv2D(
        128, 3, activation="relu", padding="same", kernel_initializer="he_normal"
    )(conv8)

    up9 = kr.layers.Conv2D(
        64, 2, activation="relu", padding="same", kernel_initializer="he_normal"
    )(kr.layers.UpSampling2D(size=(2, 2))(conv8))
    merge9 = kr.layers.Concatenate(axis=3)([conv1, up9])
    conv9 = kr.layers.Conv2D(
        64, 3, activation="relu", padding="same", kernel_initializer="he_normal"
    )(merge9)
    conv9 = kr.layers.Conv2D(
        64, 3, activation="relu", padding="same", kernel_initializer="he_normal"
    )(conv9)
    conv9 = kr.layers.Conv2D(
        2, 3, activation="relu", padding="same", kernel_initializer="he_normal"
    )(conv9)
    conv10 = kr.layers.Conv2D(nb_classes, 1, activation="sigmoid")(conv9)

    model = kr.models.Model(inputs=inputs, outputs=conv10)

    if loss_mode == "bcedice":
        loss_func = dice_coef_loss_bce
    elif loss_mode == "binary_crossentropy":
        loss_func = "binary_crossentropy"
    else:
        raise Exception(f"Unknown loss function: {loss_mode}")

    # Default learning rate for Adam: lr=1e-3, but doesn't seem to work well for unet
    # TODO: probably this needs to be removed because of the introduction of
    # the model factory
    model.compile(
        optimizer=kr.optimizers.Adam(lr=learning_rate),
        loss=loss_func,
        metrics=[
            jaccard_coef,
            jacard_coef_flat,
            jaccard_coef_int,
            dice_coef,
            "accuracy",
        ],
    )

    if init_model_weights:
        # Load the VGG16 model with pretrained weights
        # Remark: this only has 3 channels. If the input images have more channels,
        #     those channels won't be pre-loaded
        vgg = kr.applications.vgg16.VGG16(
            include_top=False,
            input_shape=(input_width, input_height, 3),
            weights="imagenet",
        )

        """
        # Create empty array with the dimensions of the first convolution layer
        conv1_weights = np.zeros((3, 3, nb_channels, 64), dtype="float32")
        conv1_weights[:, :, :3, :] = (
            vgg.get_layer("block1_conv1").get_weights()[0][:, :, :, :]
        )
        bias = vgg.get_layer("block1_conv1").get_weights()[1]
        model.get_layer('block1_conv1').set_weights((conv1_weights, bias))
        """

        # logger.info(f"VGG16 summary:")
        # vgg.summary()

        # Copy the weigths first to a new array to have flexibility in nb
        # channels of unet compared to the vgg16
        block1_conv1_weights = np.zeros((3, 3, nb_channels, 64), dtype="float32")
        vgg_block1_conv1_weights = vgg.get_layer("block1_conv1").get_weights()
        if vgg_block1_conv1_weights is None:
            raise Exception("block1_conv1_layer_weights in vgg were None")
        block1_conv1_weights[:, :, :3, :] = vgg_block1_conv1_weights[0][:, :, :, :]
        bias = vgg_block1_conv1_weights[1]
        model.get_layer("block1_conv1").set_weights((block1_conv1_weights, bias))

        block1_conv2_weights = np.zeros((3, 3, 64, 64), dtype="float32")
        vgg_block1_conv2_weights = vgg.get_layer("block1_conv2").get_weights()
        if vgg_block1_conv2_weights is None:
            raise Exception("block1_conv2_layer_weights in vgg were None")
        block1_conv2_weights[:, :, :, :] = vgg_block1_conv2_weights[0][:, :, :, :]
        bias = vgg_block1_conv2_weights[1]
        model.get_layer("block1_conv2").set_weights((block1_conv2_weights, bias))

        block2_conv1_weights = np.zeros((3, 3, 64, 128), dtype="float32")
        vgg_block2_conv1_weights = vgg.get_layer("block2_conv1").get_weights()
        if vgg_block2_conv1_weights is None:
            raise Exception("block2_conv1_layer_weights in vgg were None")
        block2_conv1_weights[:, :, :, :] = vgg_block2_conv1_weights[0][:, :, :, :]
        bias = vgg_block2_conv1_weights[1]
        model.get_layer("block2_conv1").set_weights((block2_conv1_weights, bias))

        block2_conv2_weights = np.zeros((3, 3, 128, 128), dtype="float32")
        vgg_block2_conv2_weights = vgg.get_layer("block2_conv2").get_weights()
        if vgg_block2_conv2_weights is None:
            raise Exception("block2_conv2_layer_weights in vgg were None")
        block2_conv2_weights[:, :, :, :] = vgg_block2_conv2_weights[0][:, :, :, :]
        bias = vgg_block2_conv2_weights[1]
        model.get_layer("block2_conv2").set_weights((block2_conv2_weights, bias))

    # model.summary()

    if pretrained_weights_filepath:
        """
        logger.info(f"Load model from {pretrained_weights_filepath}")

        model = kr.models.load_model(
                pretrained_weights_filepath,
                custom_objects={'jaccard_coef': jaccard_coef,
                                'jacard_coef_flat': jacard_coef_flat,
                                'jaccard_coef_int': jaccard_coef_int,
                                'dice_coef': dice_coef,
                                'dice_coef_loss_bce': dice_coef_loss_bce})

        model = get_unet(
            input_width=input_width, input_height=input_height,
            nb_channels=nb_channels, nb_classes=nb_classes, loss_mode=loss_mode,
            init_with_vgg16=False, pretrained_weights_filepath=None
        )
        model_preload = get_unet(
            input_width=input_width, input_height=input_height,
            nb_channels=nb_channels, nb_classes=nb_classes, loss_mode=loss_mode,
            init_with_vgg16=False, pretrained_weights_filepath=None
        )
        logger.info(f"Preload weights from {pretrained_weights_filepath}")
        model_preload.load_weights(pretrained_weights_filepath)

        for layer_preloaded, layer_built in zip(model_preload.layers, model.layers):
            layer_built.set_weights(layer_preloaded.get_weights())
        """

        logger.info(f"Preload weights from {pretrained_weights_filepath}")
        model.load_weights(pretrained_weights_filepath)

    return model


"""
def load_unet_model(
        filepath: str,
        learning_rate: float = None):
    logger.info(f"Load model from {filepath}")
    model = kr.models.load_model(
            filepath,
            custom_objects={'jaccard_coef': jaccard_coef,
                            'jacard_coef_flat': jacard_coef_flat,
                            'jaccard_coef_int': jaccard_coef_int,
                            'dice_coef': dice_coef,
                            'dice_coef_loss_bce': dice_coef_loss_bce})
    if learning_rate:
        kr.backend.set_value(model.optimizer.lr, learning_rate)

    return model
"""

# ------------------------------------------
# Loss functions
# ------------------------------------------


def dice_coef_loss(y_true, y_pred):
    return 1 - dice_coef(y_true, y_pred)


def bootstrapped_crossentropy(y_true, y_pred, bootstrap_type="hard", alpha=0.95):
    target_tensor = y_true
    prediction_tensor = y_pred
    _epsilon = tf.convert_to_tensor(
        kr.backend.epsilon(), prediction_tensor.dtype.base_dtype
    )
    prediction_tensor = tf.clip_by_value(prediction_tensor, _epsilon, 1 - _epsilon)  # type: ignore
    prediction_tensor = tf.math.log(prediction_tensor / (1 - prediction_tensor))  # type: ignore

    if bootstrap_type == "soft":
        bootstrap_target_tensor = alpha * target_tensor + (1.0 - alpha) * tf.sigmoid(
            prediction_tensor
        )
    else:
        bootstrap_target_tensor = alpha * target_tensor + (1.0 - alpha) * tf.cast(
            tf.sigmoid(prediction_tensor) > 0.5, tf.float32
        )
    return kr.backend.mean(
        tf.nn.sigmoid_cross_entropy_with_logits(
            labels=bootstrap_target_tensor, logits=prediction_tensor
        )
    )


def dice_coef_loss_bce(y_true, y_pred):
    dice = 0.5
    bce = 0.5
    bootstrapping = "hard"
    alpha = 1.0
    return (
        bootstrapped_crossentropy(y_true, y_pred, bootstrapping, alpha) * bce
        + dice_coef_loss(y_true, y_pred) * dice
    )


# ------------------------------------------
# Metrics functions
# ------------------------------------------

SMOOTH_LOSS = 1e-12


def jaccard_coef(y_true, y_pred):
    intersection = kr.backend.sum(y_true * y_pred, axis=[0, -1, -2])
    sum_ = kr.backend.sum(y_true + y_pred, axis=[0, -1, -2])

    jac = (intersection + SMOOTH_LOSS) / (sum_ - intersection + SMOOTH_LOSS)

    return kr.backend.mean(jac)


def jaccard_coef_int(y_true, y_pred):
    y_pred_pos = kr.backend.round(kr.backend.clip(y_pred, 0, 1))

    intersection = kr.backend.sum(y_true * y_pred_pos, axis=[0, -1, -2])
    sum_ = kr.backend.sum(y_true + y_pred_pos, axis=[0, -1, -2])
    jac = (intersection + SMOOTH_LOSS) / (sum_ - intersection + SMOOTH_LOSS)
    return kr.backend.mean(jac)


def jacard_coef_flat(y_true, y_pred):
    y_true_f = kr.backend.flatten(y_true)
    y_pred_f = kr.backend.flatten(y_pred)
    intersection = kr.backend.sum(y_true_f * y_pred_f)
    return (intersection + SMOOTH_LOSS) / (
        kr.backend.sum(y_true_f) + kr.backend.sum(y_pred_f) - intersection + SMOOTH_LOSS
    )


def dice_coef(y_true, y_pred, smooth=1.0):
    y_true_f = kr.backend.flatten(y_true)
    y_pred_f = kr.backend.flatten(y_pred)
    intersection = kr.backend.sum(y_true_f * y_pred_f)
    return (2.0 * intersection + smooth) / (
        kr.backend.sum(y_true_f) + kr.backend.sum(y_pred_f) + smooth
    )


def pct_wrong(y_true, y_pred):
    y_pred_pos = kr.backend.round(kr.backend.clip(y_pred, 0, 1))

    intersection = kr.backend.sum(y_true * y_pred_pos, axis=[0, -1, -2])
    sum_ = kr.backend.sum(y_true + y_pred_pos, axis=[0, -1, -2])
    jac = (intersection + SMOOTH_LOSS) / (sum_ - intersection + SMOOTH_LOSS)
    return kr.backend.mean(jac)



# -*- coding: utf-8 -*-
"""
High-level API to run a segmentation.
"""

import argparse
import logging
import os
from pathlib import Path
import shlex
import sys
import traceback

import geofileops as gfo

# import os
# os.environ["CUDA_VISIBLE_DEVICES"] = "-1" # Disable using GPU
import tensorflow as tf

# orthoseg is higher in dir hierarchy, add root to sys.path
sys.path.insert(0, str(Path(__file__).resolve().parent.parent))
from orthoseg.helpers import config_helper as conf
from orthoseg.helpers import email_helper
from orthoseg.lib import predicter
import orthoseg.model.model_factory as mf
import orthoseg.model.model_helper as mh
from orthoseg.util import log_util

# -------------------------------------------------------------
# First define/init general variables/constants
# -------------------------------------------------------------
# Get a logger...
logger = logging.getLogger(__name__)

# -------------------------------------------------------------
# The real work
# -------------------------------------------------------------


def predict_argstr(argstr):
    args = shlex.split(argstr)
    predict_args(args)


def predict_args(args):

    # Interprete arguments
    parser = argparse.ArgumentParser(add_help=False)

    # Required arguments
    required = parser.add_argument_group("Required arguments")
    required.add_argument(
        "-c", "--config", type=str, required=True, help="The config file to use"
    )

    # Optional arguments
    optional = parser.add_argument_group("Optional arguments")
    # Add back help
    optional.add_argument(
        "-h",
        "--help",
        action="help",
        default=argparse.SUPPRESS,
        help="Show this help message and exit",
    )

    # Interprete arguments
    args = parser.parse_args(args)

    # Run!
    predict(config_path=Path(args.config))


def predict(config_path: Path):
    """
    Run a prediction for the config specified.

    Args:
        config_path (Path): Path to the config file to use.
    """
    # Init
    # Load the config and save in a bunch of global variables zo it
    # is accessible everywhere
    conf.read_orthoseg_config(config_path)

    # Init logging
    log_util.clean_log_dir(
        log_dir=conf.dirs.getpath("log_dir"),
        nb_logfiles_tokeep=conf.logging.getint("nb_logfiles_tokeep"),
    )
    global logger
    logger = log_util.main_log_init(conf.dirs.getpath("log_dir"), __name__)

    # Log start + send email
    message = f"Start predict for config {config_path.stem}"
    logger.info(message)
    logger.debug(f"Config used: \n{conf.pformat_config()}")
    email_helper.sendmail(message)

    try:
        # Read some config, and check if values are ok
        image_layer = conf.image_layers[conf.predict["image_layer"]]
        if image_layer is None:
            raise Exception(
                f"image_layer to predict is not specified in config: {image_layer}"
            )
        input_image_dir = conf.dirs.getpath("predict_image_input_dir")
        if not input_image_dir.exists():
            raise Exception(f"input image dir doesn't exist: {input_image_dir}")

        # TODO: add something to delete old data, predictions???

        # Create base filename of model to use
        # TODO: is force data version the most logical, or rather implement
        #       force weights file or ?
        traindata_id = None
        force_model_traindata_id = conf.train.getint("force_model_traindata_id")
        if force_model_traindata_id is not None and force_model_traindata_id > -1:
            traindata_id = force_model_traindata_id

        # Get the best model that already exists for this train dataset
        trainparams_id = conf.train.getint("trainparams_id")
        best_model = mh.get_best_model(
            model_dir=conf.dirs.getpath("model_dir"),
            segment_subject=conf.general["segment_subject"],
            traindata_id=traindata_id,
            trainparams_id=trainparams_id,
        )

        # Check if a model was found
        if best_model is None:
            message = (
                f"No model found in model_dir: {conf.dirs.getpath('model_dir')} for "
                f"traindata_id: {traindata_id}"
            )
            logger.critical(message)
            raise Exception(message)
        else:
            model_weights_filepath = best_model["filepath"]
            logger.info(f"Best model found: {model_weights_filepath}")

        # Load the hyperparams of the model
        # TODO: move the hyperparams filename formatting to get_models...
        hyperparams_path = (
            best_model["filepath"].parent
            / f"{best_model['basefilename']}_hyperparams.json"
        )
        hyperparams = mh.HyperParams(path=hyperparams_path)

        # Prepare output subdir to be used for predictions
        predict_out_subdir = f"{best_model['basefilename']}"
        if trainparams_id > 0:
            predict_out_subdir += f"_{trainparams_id}"
        predict_out_subdir += f"_{best_model['epoch']}"

        # Try optimizing model with tensorrt. Not supported on Windows
        model = None
        if os.name != "nt":
            try:
                # Try import
                from tensorflow.python.compiler.tensorrt import trt_convert as trt

                # Import didn't fail, so optimize model
                logger.info(
                    "Tensorrt is available, so try to create and use optimized model"
                )
                savedmodel_optim_dir = (
                    best_model["filepath"].parent
                    / f"{best_model['filepath'].stem}_optim"
                )
                if not savedmodel_optim_dir.exists():
                    # If base model not yet in savedmodel format
                    savedmodel_dir = (
                        best_model["filepath"].parent / best_model["filepath"].stem
                    )
                    if not savedmodel_dir.exists():
                        logger.info(
                            f"SavedModel format not yet available, so load "
                            f"model + weights from {best_model['filepath']}"
                        )
                        model = mf.load_model(best_model["filepath"], compile=False)
                        logger.info(f"Now save again as savedmodel to {savedmodel_dir}")
                        tf.saved_model.save(model, str(savedmodel_dir))
                        del model

                    # Now optimize model
                    logger.info(f"Optimize + save model to {savedmodel_optim_dir}")
                    converter = trt.TrtGraphConverterV2(
                        input_saved_model_dir=str(savedmodel_dir),
                        is_dynamic_op=True,
                        precision_mode="FP16",
                    )
                    converter.convert()
                    converter.save(savedmodel_optim_dir)

                logger.info(
                    f"Load optimized model + weights from {savedmodel_optim_dir}"
                )
                model = tf.keras.models.load_model(str(savedmodel_optim_dir))

            except ImportError:
                logger.info("Tensorrt is not available, so load unoptimized model")
            except Exception as ex:
                logger.info(
                    "An error occured trying to use tensorrt, "
                    f"so load unoptimized model. Error: {ex}"
                )

        # If model isn't loaded yet... load!
        if model is None:
            model = mf.load_model(best_model["filepath"], compile=False)

        # Prepare the model for predicting
        nb_gpu = len(tf.config.experimental.list_physical_devices("GPU"))
        batch_size = conf.predict.getint("batch_size")
        if nb_gpu <= 1:
            model_for_predict = model
            logger.info(f"Predict using single GPU or CPU, with nb_gpu: {nb_gpu}")
        else:
            # If multiple GPU's available, create multi_gpu_model
            try:
                model_for_predict = model
                logger.warn("Predict using multiple GPUs NOT IMPLEMENTED AT THE MOMENT")

                # logger.info(
                #     f"Predict using multiple GPUs: {nb_gpu}, batch size becomes: "
                #     f"{batch_size*nb_gpu}"
                # )
                # batch_size *= nb_gpu
            except ValueError:
                logger.info("Predict using single GPU or CPU")
                model_for_predict = model

        # Prepare params for the inline postprocessing of the prediction
        min_probability = conf.predict.getfloat("min_probability")
        postproces = {}
        simplify_algorithm = conf.predict.get("simplify_algorithm")
        if simplify_algorithm is not None:
            postproces["simplify"] = {}
            simplify = postproces["simplify"]

            simplify_algorithm = gfo.SimplifyAlgorithm[simplify_algorithm]
            simplify["simplify_algorithm"] = simplify_algorithm
            simplify["simplify_tolerance"] = conf.predict.geteval("simplify_tolerance")
            simplify["simplify_lookahead"] = conf.predict.getint("simplify_lookahead")
            simplify["simplify_topological"] = conf.predict.getboolean_ext(
                "simplify_topological"
            )
        postproces["filter_background_modal_size"] = conf.predict.getint(
            "filter_background_modal_size"
        )
        postproces["reclassify_to_neighbour_query"] = conf.predict.get(
            "reclassify_to_neighbour_query"
        )

        # Prepare the output dirs/paths
        predict_output_dir = Path(
            f"{str(conf.dirs.getpath('predict_image_output_basedir'))}_"
            f"{predict_out_subdir}"
        )
        output_vector_dir = conf.dirs.getpath("output_vector_dir")
        output_vector_name = (
            f"{best_model['basefilename']}_{best_model['epoch']}_"
            f"{conf.predict['image_layer']}"
        )
        output_vector_path = output_vector_dir / f"{output_vector_name}.gpkg"

        # Predict for entire dataset
        nb_parallel = conf.general.getint("nb_parallel")
        predicter.predict_dir(
            model=model_for_predict,  # type: ignore
            input_image_dir=input_image_dir,
            output_image_dir=predict_output_dir,
            output_vector_path=output_vector_path,
            classes=hyperparams.architecture.classes,
            min_probability=min_probability,
            postproces=postproces,
            border_pixels_to_ignore=conf.predict.getint("image_pixels_overlap"),
            projection_if_missing=image_layer["projection"],
            input_mask_dir=None,
            batch_size=batch_size,
            evaluate_mode=False,
            cancel_filepath=conf.files.getpath("cancel_filepath"),
            nb_parallel_postprocess=nb_parallel,
        )

        # Log and send mail
        message = f"Completed predict for config {config_path.stem}"
        logger.info(message)
        email_helper.sendmail(message)
    except Exception as ex:
        message = f"ERROR while running predict for task {config_path.stem}"
        logger.exception(message)
        email_helper.sendmail(
            subject=message, body=f"Exception: {ex}\n\n {traceback.format_exc()}"
        )
        raise Exception(message) from ex


def main():
    try:
        predict_args(sys.argv[1:])
    except Exception as ex:
        logger.exception(f"Error: {ex}")
        raise


# If the script is ran directly...
if __name__ == "__main__":
    main()



#!/usr/bin/python3
# Based on https://github.com/sdushantha/gitdir/blob/master/gitdir/gitdir.py

import re
import os
import ssl
from typing import Union
import urllib.request
import json
from pathlib import Path


def create_url(url):
    """
    From the given url, produce a URL that is compatible with Github's REST API.
    Can handle blob or tree paths.
    """

    # extract the branch name from the given url (e.g master)
    branch = re.findall(r"/tree/(.*?)/", url)
    api_url = url.replace("https://github.com", "https://api.github.com/repos")
    if len(branch) == 0:
        branch = re.findall(r"/blob/(.*?)/", url)[0]
        download_dirs = re.findall(r"/blob/" + branch + r"/(.*)", url)[0]
        api_url = re.sub(r"/blob/.*?/", "/contents/", api_url)
    else:
        branch = branch[0]
        download_dirs = re.findall(r"/tree/" + branch + r"/(.*)", url)[0]
        api_url = re.sub(r"/tree/.*?/", "/contents/", api_url)

    api_url = api_url + "?ref=" + branch
    return api_url, download_dirs


def download(
    repo_url: str, output_dir: Path, ssl_verify: Union[bool, str, None] = None
):
    """
    Downloads the files and directories in repo_url.

    Args
        repo_url (str): url to the repository to download.
        output_dir (Path): directory to download the repository to.
        ssl_verify (bool or str, optional): True or None to use the default
            certificate bundle as installed on your system. False disables
            certificate validation (NOT recommended!). If a path to a
            certificate bundle file (.pem) is passed, this will be used.
            In corporate networks using a proxy server this is often needed
            to evade CERTIFICATE_VERIFY_FAILED errors. Defaults to None.
    """
    context = None
    if ssl_verify is not None:
        # If it is a string, make sure it isn't actually a bool
        if isinstance(ssl_verify, str):
            if ssl_verify.lower() == "true":
                context = None
            elif ssl_verify.lower() == "false":
                ssl_verify = False
            else:
                context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
                context.load_verify_locations(ssl_verify)

        if isinstance(ssl_verify, bool) and not ssl_verify:
            context = ssl._create_unverified_context()
            print("SSL VERIFICATION IS TURNED OFF!!!")

    # generate the url which returns the JSON data
    api_url, download_dirs = create_url(repo_url)

    # To handle file names.
    if len(download_dirs.split(".")) == 0:
        dir_out = output_dir / download_dirs
    else:
        dir_out = output_dir / "/".join(download_dirs.split("/")[0:-1])

    # Download the file
    with urllib.request.urlopen(api_url, context=context) as u:
        # make a directory with the name which is taken from
        # the actual repo
        dir_out.mkdir(parents=True, exist_ok=True)

        # total files count
        total_files = 0
        raw_data = u.read()
        data = json.loads(raw_data)

        # Get the total number of files
        total_files += len(data)

        # If the data is a file, download it as one.
        if isinstance(data, dict) and data["type"] == "file":
            # download the file
            with (
                urllib.request.urlopen(data["download_url"], context=context) as u,
                open(dir_out / data["name"], "wb") as f,
            ):
                f.write(u.read())
            return total_files

        # Loop over all files/dirs found
        for file in data:
            file_url = file["download_url"]
            path = output_dir / file["path"]
            os.makedirs(path.parent, exist_ok=True)

            # If it is a file, download it, if dir, start recursively
            if file_url is not None:
                with (
                    urllib.request.urlopen(file_url, context=context) as u,
                    open(path, "wb") as f,
                ):
                    f.write(u.read())
            else:
                download(file["html_url"], output_dir)

    return total_files



# -*- coding: utf-8 -*-
"""
Tests for functionalities in orthoseg.train.
"""

from pathlib import Path
import shutil
import sys

import geofileops as gfo

# Add path so the local orthoseg packages are found
sys.path.insert(0, str(Path(__file__).resolve().parent.parent))
import orthoseg

from orthoseg.helpers import config_helper as conf
import orthoseg.model.model_helper as mh
from orthoseg.util import gdrive_util

# ----------------------------------------------------
# Tests
# ----------------------------------------------------


def get_testdata_dir() -> Path:
    return Path(__file__).resolve().parent / "data"


def get_test_projects_dir() -> Path:
    return Path(__file__).resolve().parent / "test_projects"


def test_load_images():
    # Load project config to init some vars.
    config_path = (
        get_test_projects_dir() / "footballfields" / "footballfields_BEFL-2019.ini"
    )
    conf.read_orthoseg_config(config_path)
    image_cache_dir = conf.dirs.getpath("predict_image_input_dir")

    # Clean result if it isn't empty yet
    if image_cache_dir.exists():
        shutil.rmtree(image_cache_dir)
        # Make sure is is deleted now!
        assert image_cache_dir.exists() is False

    # Run task to load images
    orthoseg.load_images(config_path)

    # Check if the right number of files was loaded
    assert image_cache_dir.exists() is True
    files = list(image_cache_dir.glob("**/*.jpg"))
    assert len(files) == 8


def test_train():
    # Load project config to init some vars.
    config_path = (
        get_test_projects_dir() / "footballfields" / "footballfields_train.ini"
    )
    conf.read_orthoseg_config(config_path)

    # Init + cleanup result dirs
    traindata_id_result = 2
    training_dir = conf.dirs.getpath("training_dir")
    training_id_dir = training_dir / f"{traindata_id_result:02d}"
    if training_id_dir.exists():
        shutil.rmtree(training_id_dir)
    model_dir = conf.dirs.getpath("model_dir")
    if model_dir.exists():
        modelfile_paths = model_dir.glob(f"footballfields_{traindata_id_result:02d}_*")
        for modelfile_path in modelfile_paths:
            modelfile_path.unlink()

    # Download the version 01 model
    model_hdf5_path = model_dir / "footballfields_01_0.92512_242.hdf5"
    if model_hdf5_path.exists() is False:
        gdrive_util.download_file("1XmAenCW6K_RVwqC6xbkapJ5ws-f7-QgH", model_hdf5_path)
    model_hyperparams_path = model_dir / "footballfields_01_hyperparams.json"
    if model_hyperparams_path.exists() is False:
        gdrive_util.download_file(
            "1umxcd4RkB81sem9PdIpLoWeiIW8ga1u7", model_hyperparams_path
        )
    model_modeljson_path = model_dir / "footballfields_01_model.json"
    if model_modeljson_path.exists() is False:
        gdrive_util.download_file(
            "16qe8thBTrO3dFfLMU1T22gWcfHVXt8zQ", model_modeljson_path
        )

    # Run train session
    # The label files are newer than the ones used to train the current model,
    # so a new model will be trained.
    orthoseg.train(config_path)

    # Check if the training (image) data was created
    assert training_id_dir.exists() is True

    # Check if the new model was created
    best_model = mh.get_best_model(
        model_dir=conf.dirs.getpath("model_dir"),
        segment_subject=conf.general["segment_subject"],
        traindata_id=2,
    )

    assert best_model is not None
    assert best_model["traindata_id"] == traindata_id_result
    assert best_model["epoch"] == 0


def test_predict():
    # Load project config to init some vars.
    config_path = (
        get_test_projects_dir() / "footballfields" / "footballfields_BEFL-2019.ini"
    )
    conf.read_orthoseg_config(config_path)

    # Cleanup result if it isn't empty yet
    predict_image_output_basedir = conf.dirs.getpath("predict_image_output_basedir")
    predict_image_output_dir = (
        predict_image_output_basedir.parent
        / f"{predict_image_output_basedir.name}_footballfields_02_0"
    )
    if predict_image_output_dir.exists():
        shutil.rmtree(predict_image_output_dir)
        # Make sure it is deleted now!
        assert predict_image_output_dir.exists() is False
    result_vector_dir = conf.dirs.getpath("output_vector_dir")
    if result_vector_dir.exists():
        shutil.rmtree(result_vector_dir)
        # Make sure is is deleted now!
        assert result_vector_dir.exists() is False

    # Run task to predict
    orthoseg.predict(config_path)

    # Check results
    result_vector_path = result_vector_dir / "footballfields_01_242_BEFL-2019.gpkg"
    assert result_vector_path.exists() is True
    result_gdf = gfo.read_file(result_vector_path)
    assert len(result_gdf) == 11


def test_postprocess():
    # Load project config to init some vars.
    config_path = (
        get_test_projects_dir() / "footballfields" / "footballfields_BEFL-2019.ini"
    )
    conf.read_orthoseg_config(config_path)

    # Cleanup result if it isn't empty yet
    result_vector_dir = conf.dirs.getpath("output_vector_dir")
    result_diss_path = (
        result_vector_dir / "footballfields_01_242_BEFL-2019_dissolve.gpkg"
    )
    result_simpl_path = (
        result_vector_dir / "footballfields_01_242_BEFL-2019_dissolve_simpl.gpkg"
    )
    if result_diss_path.exists():
        gfo.remove(result_diss_path)
    if result_simpl_path.exists():
        gfo.remove(result_simpl_path)

    # Run task to postprocess
    orthoseg.postprocess(config_path)

    # Check results
    assert result_diss_path.exists() is True
    result_gdf = gfo.read_file(result_diss_path)
    assert len(result_gdf) == 10

    assert result_simpl_path.exists() is True
    result_gdf = gfo.read_file(result_simpl_path)
    assert len(result_gdf) == 10


if __name__ == "__main__":
    test_load_images()
    test_train()
    test_predict()
    test_postprocess()



# -*- coding: utf-8 -*-
"""
Modile with generic Utility functions for vector manipulations.
"""

import copy
import logging
from typing import Optional, Tuple

import geofileops as gfo
from geofileops.util import geometry_util
from geofileops.util import geoseries_util
import geopandas as gpd
import numpy as np
import pandas as pd
import pygeos
from shapely import geometry as sh_geom
import topojson
import topojson.ops

# -------------------------------------------------------------
# First define/init some general variables/constants
# -------------------------------------------------------------
logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)

# -------------------------------------------------------------
# The real work
# -------------------------------------------------------------


def calc_onborder(
    geoms_gdf: gpd.GeoDataFrame,
    border_bounds: Tuple[float, float, float, float],
    onborder_column_name: str = "onborder",
) -> gpd.GeoDataFrame:
    """
    Add/update a column to the GeoDataFrame with:
        * 0 if the polygon isn't on the border and
        * 1 if it is.

    Args
        geoms_gdf: input GeoDataFrame
        border_bounds: the bounds (tupple with (xmin, ymin, xmax, ymax)
                       to check against to determine onborder
        onborder_column_name: the column name of the onborder column

    """
    # Split geoms that need unioning versus geoms that don't
    # -> They are on the edge of a tile
    if geoms_gdf is not None and len(geoms_gdf.index) > 0:

        # Check
        for i, geom_row in geoms_gdf.iterrows():
            # Check if the geom is on the border of the tile
            geom_bounds = geom_row["geometry"].bounds  # type: ignore
            onborder = 0
            if (
                geom_bounds[0] <= border_bounds[0]
                or geom_bounds[1] <= border_bounds[1]
                or geom_bounds[2] >= border_bounds[2]
                or geom_bounds[3] >= border_bounds[3]
            ):
                onborder = 1

            geoms_gdf.loc[i, onborder_column_name] = onborder  # type: ignore

    geoms_gdf.reset_index(drop=True, inplace=True)
    return geoms_gdf


def is_valid_reason(geoseries) -> pd.Series:
    # Remark: should be moved to geofileops (till available in geopandas)!!!
    return pd.Series(
        data=pygeos.is_valid_reason(geoseries.array.data), index=geoseries.index
    )  # type: ignore


def simplify_topo_orthoseg(
    geoseries: gpd.GeoSeries,
    tolerance: float,
    algorithm: gfo.SimplifyAlgorithm = gfo.SimplifyAlgorithm.RAMER_DOUGLAS_PEUCKER,
    lookahead: int = 8,
    keep_points_on: Optional[sh_geom.base.BaseGeometry] = None,
) -> gpd.GeoSeries:
    """
    Applies simplify while retaining common boundaries between geometries in the
    geoseries.

    This is a specific version for orthoseg that, when algorithm is LANG, applies rdp
    simplify with halve the tolerance first.

    Args:
        geoseries (gpd.GeoSeries): the geoseries to simplify.
        algorithm (SimplifyAlgorithm): algorithm to use.
        tolerance (float): tolerance to use for simplify
        lookahead (int, optional): lookahead value for algorithms that use this.
            Defaults to 8.
        keep_points_on (Optional[sh_geom.base.BaseGeometry], optional): points that
            intersect with this geometry won't be removed by the simplification.
            Defaults to None.

    Returns:
        gpd.GeoSeries: the simplified geoseries
    """
    # Just return empty geoseries
    if len(geoseries) == 0:
        return geoseries

    # Copy geoseries
    geoseries_copy = geoseries.copy()  # type: ignore

    # Set empty geometries to None, if no geometries are left, return
    empty_idxs = pygeos.is_empty(geoseries_copy.array.data).nonzero()
    if len(empty_idxs) > 0:
        geoseries_copy.iloc[empty_idxs] = None
    if len(geoseries_copy[geoseries_copy != np.array(None)]) == 0:
        return geoseries

    topo = topojson.Topology(geoseries_copy, prequantize=False, shared_coords=False)
    # 46889.5, 211427.5
    # 46932.5, 211391.5

    def simplify_topolines(
        _topolines: list,
        _algorithm: gfo.SimplifyAlgorithm,
        _tolerance: float,
        _lookahead: int,
        _keep_points_on: Optional[sh_geom.base.BaseGeometry],
    ) -> list:
        """
        Internal function that takes care of some plumbing specific to simplifying
        the lines in the topologies.

        Args:
            _topolines (list): _description_
            _algorithm (gfo.SimplifyAlgorithm): _description_
            _tolerance (float): _description_
            _lookahead (int): _description_
            _keep_points_on (Optional[sh_geom.base.BaseGeometry]): _description_

        Returns:
            list: _description_
        """
        topolines_geom = sh_geom.MultiLineString(_topolines)
        topolines_simpl = geometry_util.simplify_ext(
            geometry=topolines_geom,
            tolerance=_tolerance,
            algorithm=_algorithm,
            lookahead=_lookahead,
            keep_points_on=_keep_points_on,
            preserve_topology=True,
        )
        assert topolines_simpl is not None

        # Copy the results of the simplified lines
        if _algorithm == gfo.SimplifyAlgorithm.LANG:
            # For LANG, a simple copy is OK
            if isinstance(topolines_simpl, sh_geom.LineString):
                return [list(topolines_simpl.coords)]
            else:
                assert isinstance(topolines_simpl, sh_geom.MultiLineString)
                return [list(geom.coords) for geom in topolines_simpl.geoms]
        else:
            # For RDP, only overwrite the lines that have a valid result
            topolines_copy = copy.deepcopy(_topolines)
            for index in range(len(topolines_copy)):
                # Get the coordinates of the simplified version
                if isinstance(topolines_simpl, sh_geom.base.BaseMultipartGeometry):
                    topoline_simpl = topolines_simpl.geoms[index].coords  # type: ignore
                else:
                    topoline_simpl = topolines_simpl.coords
                # If the result of the simplify is a point, keep original
                if len(topoline_simpl) < 2:
                    continue
                elif (
                    list(topoline_simpl[0]) != topolines_copy[index][0]
                    or list(topoline_simpl[-1]) != topolines_copy[index][-1]
                ):
                    # Start or end point of the simplified version is not the same
                    # anymore
                    continue
                else:
                    topolines_copy[index] = list(topoline_simpl)

            return topolines_copy

    if algorithm == gfo.SimplifyAlgorithm.LANG:
        # For lang, first pre-simplify using rdp
        topolines_simplified = simplify_topolines(
            _topolines=topo.output["arcs"],
            _algorithm=gfo.SimplifyAlgorithm.RAMER_DOUGLAS_PEUCKER,
            _tolerance=tolerance / 2,
            _lookahead=lookahead,
            _keep_points_on=keep_points_on,
        )
        for line_id in range(len(topo.output["arcs"])):
            topo.output["arcs"][line_id] = topolines_simplified[line_id]

        topolines_simplified = simplify_topolines(
            _topolines=topo.output["arcs"],
            _algorithm=algorithm,
            _tolerance=tolerance,
            _lookahead=lookahead,
            _keep_points_on=keep_points_on,
        )
        for line_id in range(len(topo.output["arcs"])):
            topo.output["arcs"][line_id] = topolines_simplified[line_id]
    else:
        topolines_simplified = simplify_topolines(
            _topolines=topo.output["arcs"],
            _algorithm=algorithm,
            _tolerance=tolerance,
            _lookahead=lookahead,
            _keep_points_on=keep_points_on,
        )
        for line_id in range(len(topo.output["arcs"])):
            topo.output["arcs"][line_id] = topolines_simplified[line_id]

    topo_simpl_geoseries = topo.to_gdf(crs=geoseries.crs).geometry
    assert isinstance(topo_simpl_geoseries, gpd.GeoSeries)
    topo_simpl_geoseries.array.data = pygeos.make_valid(  # type: ignore
        topo_simpl_geoseries.array.data  # type: ignore
    )
    geometry_types_orig = geoseries.type.unique()
    geometry_types_orig = geometry_types_orig[geometry_types_orig != None]
    geometry_types_simpl = topo_simpl_geoseries.type.unique()
    if len(geometry_types_orig) == 1 and (
        len(geometry_types_simpl) > 1
        or geometry_types_orig[0] != geometry_types_simpl[0]
    ):
        topo_simpl_geoseries = geoseries_util.geometry_collection_extract(
            topo_simpl_geoseries,
            gfo.GeometryType(geometry_types_orig[0]).to_primitivetype,
        )
    return topo_simpl_geoseries



# -*- coding: utf-8 -*-
"""
Module with specific helper functions to manage the configuration of orthoseg.
"""

import configparser
import json
import logging
from pathlib import Path
import pprint

from orthoseg.util import config_util

# -------------------------------------------------------------
# First define/init some general variables/constants
# -------------------------------------------------------------
# Get a logger...
logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)

# Define the chars that cannot be used in codes that are use in filenames.
# Remark: '_' cannot be used because '_' is used as devider to parse filenames, and if
# it is used in codes as well the parsing becomes a lot more difficult.
illegal_chars_in_codes = ["_", ",", ".", "?", ":"]

# -------------------------------------------------------------
# The real work
# -------------------------------------------------------------


def read_orthoseg_config(config_path: Path):

    # Determine list of config files that should be loaded
    config_paths = config_util.get_config_files(config_path)  # type: ignore
    # Load them
    global config
    config = config_util.read_config_ext(config_paths)

    # Now do orthoseg-specific checks, inits,... on config
    global config_filepaths_used
    config_filepaths_used = config_paths

    # Now set global variables to each section as shortcuts
    global general
    general = config["general"]
    global model
    model = config["model"]
    global download
    download = config["download"]
    global train
    train = config["train"]
    global predict
    predict = config["predict"]
    global postprocess
    postprocess = config["postprocess"]
    global dirs
    dirs = config["dirs"]
    global files
    files = config["files"]
    global logging
    logging = config["logging"]
    global email
    email = config["email"]

    # Some checks to make sure the config is loaded properly
    segment_subject = general.get("segment_subject")
    if segment_subject is None or segment_subject == "MUST_OVERRIDE":
        raise Exception(
            "Projectconfig parameter general.segment_subject needs to be overruled to"
            "a proper name in a specific project config file, \nwith config_filepaths "
            f"{config_paths}"
        )
    elif any(
        illegal_character in segment_subject
        for illegal_character in illegal_chars_in_codes
    ):
        raise Exception(
            f"Projectconfig parameter general.segment_subject ({segment_subject}) "
            f"should not contain any of the following chars: {illegal_chars_in_codes}"
        )

    # If the projects_dir parameter is a relative path, resolve it towards the location
    # of the project config file.
    projects_dir = dirs.getpath("projects_dir")
    if not projects_dir.is_absolute():
        projects_dir_absolute = (config_paths[-1].parent / projects_dir).resolve()
        logger.info(
            f"dirs.projects_dir was relative: is resolved to {projects_dir_absolute}"
        )
        dirs["projects_dir"] = projects_dir_absolute.as_posix()

    # Read the layer config
    layer_config_filepath = files.getpath("image_layers_config_filepath")
    global layer_config_filepath_used
    layer_config_filepath_used = layer_config_filepath

    global image_layers
    image_layers = read_layer_config(layer_config_filepath=layer_config_filepath)


def read_layer_config(layer_config_filepath: Path) -> dict:
    # Init
    if not layer_config_filepath.exists():
        raise Exception(f"Layer config file not found: {layer_config_filepath}")

    # Read config file...
    layer_config = configparser.ConfigParser(
        interpolation=configparser.ExtendedInterpolation(),
        converters={
            "list": lambda x: [i.strip() for i in x.split(",")],
            "listint": lambda x: [int(i.strip()) for i in x.split(",")],
            "dict": lambda x: None if x is None else json.loads(x),
            "path": lambda x: Path(x),
        },
    )
    layer_config.read(layer_config_filepath)

    # Prepare data
    image_layers = {}
    for image_layer in layer_config.sections():
        # First check if the image_layer code doesn't contain 'illegal' characters
        if any(illegal_char in image_layer for illegal_char in illegal_chars_in_codes):
            raise Exception(
                f"Section name [{image_layer}] in layer config should not contain any"
                f"of these chars: {illegal_chars_in_codes}, in {layer_config_filepath}"
            )

        # Init layer with all parameters in the section as dict
        image_layers[image_layer] = dict(layer_config[image_layer])

        # If the layer source(s) are specified in a json parameter, parse it
        if "layersources" in image_layers[image_layer]:
            image_layers[image_layer]["layersources"] = layer_config[
                image_layer
            ].getdict("layersources")

            # Give default values to some optional properties of a server
            for layersource in image_layers[image_layer]["layersources"]:
                if "random_sleep" not in layersource:
                    layersource["random_sleep"] = 0
                if "wms_ignore_capabilities_url" not in layersource:
                    layersource["wms_ignore_capabilities_url"] = False

        else:
            # If not, the layersource should be specified in seperate parameters
            layersource = {}
            layersource["wms_server_url"] = layer_config[image_layer].get(
                "wms_server_url"
            )
            layersource["wms_version"] = layer_config[image_layer].get(
                "wms_version", fallback="1.3.0"
            )
            # The layer names and layer styles are lists
            layersource["layernames"] = layer_config[image_layer].getlist(
                "wms_layernames"
            )
            layersource["layerstyles"] = layer_config[image_layer].getlist(
                "wms_layerstyles"
            )
            # Some more properties
            layersource["bands"] = layer_config[image_layer].getlist(
                "bands", fallback=None
            )
            layersource["random_sleep"] = layer_config[image_layer].getint(
                "random_sleep", fallback=0
            )
            layersource["wms_ignore_capabilities_url"] = layer_config[
                image_layer
            ].getboolean("wms_ignore_capabilities_url", fallback=False)
            image_layers[image_layer]["layersources"] = [layersource]

        # Read nb_concurrent calls param
        image_layers[image_layer]["nb_concurrent_calls"] = layer_config[
            image_layer
        ].getint("nb_concurrent_calls", fallback=6)

        # Check if a region of interest is specified as file or bbox
        roi_filepath = layer_config[image_layer].getpath("roi_filepath", fallback=None)
        image_layers[image_layer]["roi_filepath"] = roi_filepath
        bbox_tuple = None
        if layer_config.has_option(image_layer, "bbox"):
            bbox_list = layer_config[image_layer].getlist("bbox")
            bbox_tuple = (
                float(bbox_list[0]),
                float(bbox_list[1]),
                float(bbox_list[2]),
                float(bbox_list[3]),
            )
            image_layers[image_layer]["bbox"] = bbox_tuple
        image_layers[image_layer]["bbox"] = bbox_tuple

        # Check if the grid xmin and xmax are specified
        grid_xmin = 0
        if layer_config.has_option(image_layer, "grid_xmin"):
            grid_xmin = layer_config[image_layer].getfloat("grid_xmin")
        image_layers[image_layer]["grid_xmin"] = grid_xmin
        grid_ymin = 0
        if layer_config.has_option(image_layer, "grid_ymin"):
            grid_ymin = layer_config[image_layer].getfloat("grid_ymin")
        image_layers[image_layer]["grid_ymin"] = grid_ymin

        # Check if a image_pixels_ignore_border is specified
        image_pixels_ignore_border = layer_config[image_layer].getint(
            "image_pixels_ignore_border", fallback=0
        )
        image_layers[image_layer][
            "image_pixels_ignore_border"
        ] = image_pixels_ignore_border
    return image_layers


def pformat_config():
    message = f"Config files used: {pprint.pformat(config_filepaths_used)} \n"
    message += f"Layer config file used: {layer_config_filepath_used} \n"
    message += "Config info listing:\n"
    message += pprint.pformat(config_util.as_dict(config))
    message += "Layer config info listing:\n"
    message += pprint.pformat(image_layers)
    return message


# If the script is ran directly...
if __name__ == "__main__":
    raise Exception("Not implemented")



# -*- coding: utf-8 -*-
"""
Tests for functionalities in orthoseg.lib.postprocess_predictions.
"""
import os
from pathlib import Path
import sys
from typing import List, Optional, Union

import pytest
from shapely import geometry as sh_geom

# Make hdf5 version warning non-blocking
os.environ["HDF5_DISABLE_VERSION_CHECK"] = "1"

import geopandas as gpd
import geofileops as gfo

# Add path so the local orthoseg packages are found
sys.path.insert(0, str(Path(__file__).resolve().parent.parent))
from orthoseg.helpers import config_helper
from orthoseg.lib import prepare_traindatasets as prep_traindata
from orthoseg.lib.prepare_traindatasets import ValidationError
from tests.test_helper import TestData


# Helper functions to prepare test data
# -------------------------------------


def _prepare_locations_file(
    tmp_path, locations: Optional[Union[dict, gpd.GeoDataFrame]]
) -> Path:
    locations_path = tmp_path / "locations.gpkg"
    if locations is None:
        locations = TestData.locations_gdf
    if isinstance(locations, dict):
        locations = gpd.GeoDataFrame(locations, crs="EPSG:31370")  # type: ignore

    gfo.to_file(locations, locations_path)
    return locations_path


def _prepare_polygons_file(
    tmp_path, polygons: Optional[Union[dict, gpd.GeoDataFrame]]
) -> Path:
    polygons_path = tmp_path / "polygons.gpkg"
    if polygons is None:
        polygons = TestData.polygons_gdf
    if isinstance(polygons, dict):
        polygons = gpd.GeoDataFrame(polygons, crs="EPSG:31370")  # type: ignore

    gfo.to_file(polygons, polygons_path)
    return polygons_path


def _prepare_labelinfos(
    tmp_path,
    locations: Optional[Union[dict, gpd.GeoDataFrame]] = None,
    polygons: Optional[Union[dict, gpd.GeoDataFrame]] = None,
) -> List[prep_traindata.LabelInfo]:
    locations_path = _prepare_locations_file(tmp_path, locations)
    polygons_path = _prepare_polygons_file(tmp_path, polygons)

    label_info = prep_traindata.LabelInfo(
        locations_path=locations_path,
        polygons_path=polygons_path,
        image_layer="BEFL-2019",
    )
    return [label_info]


# Actual tests
# ------------


@pytest.mark.parametrize(
    "geometry, traindata_type, expected_len_locations",
    [
        [TestData.location, "train", 4],
        [None, "validation", 2],
        [sh_geom.Polygon(), "validation", 2],
    ],
)
def test_prepare_labeldata_locations(geometry, traindata_type, expected_len_locations):
    # Prepare test data
    # Add some extra data to make sure it also works for multiple rows
    # Prepare test data, by wrapping the parametrized invalid test data by proper data
    # to make sure the checks work on multiple rows,...
    locations_gdf = gpd.GeoDataFrame(
        data={
            "geometry": [TestData.location, geometry, geometry, TestData.location],
            "traindata_type": ["train", traindata_type, traindata_type, "validation"],
            "path": "/tmp/locations.gdf",
        },
        crs="EPSG:31370",  # type: ignore
    )
    label_info = prep_traindata.LabelInfo(
        locations_path=Path("locations"),
        polygons_path=Path("polygons"),
        image_layer="BEFL-2019",
        locations_gdf=locations_gdf,
        polygons_gdf=TestData.polygons_gdf,
    )

    # Test!
    labeldata = prep_traindata.prepare_labeldata(
        label_infos=[label_info],
        classes=TestData.classes,
        labelname_column="classname",
        image_pixel_x_size=TestData.image_pixel_x_size,
        image_pixel_y_size=TestData.image_pixel_y_size,
        image_pixel_width=TestData.image_pixel_width,
        image_pixel_height=TestData.image_pixel_height,
    )
    labellocations_gdf, labelpolygons_gdf = labeldata[0]

    assert len(labelpolygons_gdf) == len(TestData.polygons_gdf)
    assert len(labellocations_gdf) == expected_len_locations


@pytest.mark.parametrize(
    "expected_error, geometry, traindata_type",
    [
        ["Invalid geometry ", TestData.location_invalid, "train"],
        ["Invalid traindata_type ", TestData.location, "invalid_traindata_type"],
        ["Location geometry too small ", TestData.location.buffer(-5), "train"],
    ],
)
def test_prepare_labeldata_locations_invalid(expected_error, geometry, traindata_type):
    # Prepare test data, by wrapping the parametrized invalid test data by proper data
    # to make sure the checks work on multiple rows,...
    locations_gdf = gpd.GeoDataFrame(
        data={
            "geometry": [TestData.polygon, geometry, geometry, TestData.polygon],
            "traindata_type": ["train", traindata_type, traindata_type, "validation"],
            "path": "/tmp/locations.gdf",
        },
        crs="EPSG:31370",  # type: ignore
    )
    label_info = prep_traindata.LabelInfo(
        locations_path=Path("locations"),
        polygons_path=Path("polygons"),
        image_layer="BEFL-2019",
        locations_gdf=locations_gdf,
        polygons_gdf=TestData.polygons_gdf,
    )

    # Test!
    with pytest.raises(ValidationError, match="Errors found in label data") as ex:
        _ = prep_traindata.prepare_labeldata(
            label_infos=[label_info],
            classes=TestData.classes,
            labelname_column="classname",
            image_pixel_x_size=TestData.image_pixel_x_size,
            image_pixel_y_size=TestData.image_pixel_y_size,
            image_pixel_width=TestData.image_pixel_width,
            image_pixel_height=TestData.image_pixel_height,
        )

    assert ex.value.errors is not None
    assert len(ex.value.errors) == 2
    assert ex.value.errors[0].startswith(expected_error)
    assert ex.value.errors[1].startswith(expected_error)


@pytest.mark.parametrize(
    "geometry, classname, expected_len_polygons",
    [
        [TestData.polygon, "testlabel1", 4],
        [None, "testlabel1", 2],
        [sh_geom.Polygon(), "testlabel1", 2],
    ],
)
def test_prepare_labeldata_polygons(geometry, classname, expected_len_polygons):
    # Prepare test data
    # Add some extra data to make sure it also works for multiple rows
    polygons_gdf = gpd.GeoDataFrame(
        data={
            "geometry": [TestData.polygon, geometry, geometry, TestData.polygon],
            "classname": ["testlabel1", classname, classname, "testlabel2"],
            "path": "/tmp/polygons.gdf",
        },
        crs="EPSG:31370",  # type: ignore
    )
    label_info = prep_traindata.LabelInfo(
        locations_path=Path("locations"),
        polygons_path=Path("polygons"),
        image_layer="BEFL-2019",
        locations_gdf=TestData.locations_gdf,
        polygons_gdf=polygons_gdf,
    )

    # Test!
    labeldata = prep_traindata.prepare_labeldata(
        label_infos=[label_info],
        classes=TestData.classes,
        labelname_column="classname",
        image_pixel_x_size=TestData.image_pixel_x_size,
        image_pixel_y_size=TestData.image_pixel_y_size,
        image_pixel_width=TestData.image_pixel_width,
        image_pixel_height=TestData.image_pixel_height,
    )
    labellocations_gdf, labelpolygons_gdf = labeldata[0]

    assert len(labelpolygons_gdf) == expected_len_polygons
    assert len(labellocations_gdf) == len(TestData.locations_gdf)


@pytest.mark.parametrize(
    "expected_error, geometry, classname",
    [
        ["Invalid geometry ", TestData.polygon_invalid, "testlabel1"],
        ["Invalid classname ", TestData.polygon, "unknown"],
    ],
)
def test_prepare_labeldata_polygons_invalid(expected_error, geometry, classname):
    # Prepare test data, by wrapping the parametrized invalid test data by proper data
    # to make sure the checks work on multiple rows,...
    polygons_gdf = gpd.GeoDataFrame(
        data={
            "geometry": [TestData.polygon, geometry, geometry, TestData.polygon],
            "classname": ["testlabel1", classname, classname, "testlabel2"],
            "path": "/tmp/polygons.gdf",
        },
        crs="EPSG:31370",  # type: ignore
    )
    label_info = prep_traindata.LabelInfo(
        locations_path=Path("locations"),
        polygons_path=Path("polygons"),
        image_layer="BEFL-2019",
        locations_gdf=TestData.locations_gdf,
        polygons_gdf=polygons_gdf,
    )

    # Test!
    with pytest.raises(ValidationError, match="Errors found in label data") as ex:
        _ = prep_traindata.prepare_labeldata(
            label_infos=[label_info],
            classes=TestData.classes,
            labelname_column="classname",
            image_pixel_x_size=TestData.image_pixel_x_size,
            image_pixel_y_size=TestData.image_pixel_y_size,
            image_pixel_width=TestData.image_pixel_width,
            image_pixel_height=TestData.image_pixel_height,
        )

    assert ex.value.errors is not None
    assert len(ex.value.errors) == 2
    assert ex.value.errors[0].startswith(expected_error)
    assert ex.value.errors[1].startswith(expected_error)


def test_prepare_labeldata_polygons_columnname_backw_compat(tmp_path):
    # Test bacwards compatibility for old label column name
    # Prepare test data
    polygons_gdf = gpd.GeoDataFrame(
        data={
            "geometry": [TestData.polygon, TestData.polygon],
            "label_name": ["testlabel1", "testlabel2"],
            "path": "/tmp/polygons.gdf",
        },
        crs="EPSG:31370",  # type: ignore
    )
    label_info = prep_traindata.LabelInfo(
        locations_path=Path("locations"),
        polygons_path=Path("polygons"),
        image_layer="BEFL-2019",
        locations_gdf=TestData.locations_gdf,
        polygons_gdf=polygons_gdf,
    )

    # Test!
    labeldata = prep_traindata.prepare_labeldata(
        label_infos=[label_info],
        classes=TestData.classes,
        labelname_column="test_columnname",
        image_pixel_x_size=TestData.image_pixel_x_size,
        image_pixel_y_size=TestData.image_pixel_y_size,
        image_pixel_width=TestData.image_pixel_width,
        image_pixel_height=TestData.image_pixel_height,
    )
    locations_gdf, polygons_to_burn_gdf = labeldata[0]

    assert len(polygons_to_burn_gdf) == 2


def test_prepare_traindata_full(tmp_path):
    # Prepare test data
    classes = TestData.classes
    image_layers_config_path = TestData.testprojects_dir / "imagelayers.ini"
    image_layers = config_helper.read_layer_config(image_layers_config_path)
    label_infos = _prepare_labelinfos(tmp_path)

    # Test with the default data...
    training_dir = tmp_path / "training_dir"
    training_dir, traindata_id = prep_traindata.prepare_traindatasets(
        label_infos=label_infos,
        classes=classes,
        image_layers=image_layers,
        training_dir=training_dir,
        image_pixel_x_size=TestData.image_pixel_x_size,
        image_pixel_y_size=TestData.image_pixel_y_size,
        image_pixel_width=TestData.image_pixel_width,
        image_pixel_height=TestData.image_pixel_height,
    )

    assert training_dir.exists() is True



# Import the high level API
# flake8: noqa: F401
from orthoseg.load_images import load_images  
from orthoseg.train import train
from orthoseg.train import _search_label_files
from orthoseg.predict import predict
from orthoseg.postprocess import postprocess



# -*- coding: utf-8 -*-
"""
Module with specific helper functions to manage progress reporting.
"""

import datetime
import logging
from typing import Optional

# -------------------------------------------------------------
# First define/init general variables/constants
# -------------------------------------------------------------
# Get a logger...
logger = logging.getLogger(__name__)

# -------------------------------------------------------------
# The real work
# -------------------------------------------------------------


class ProgressHelper:
    first_reporting_done = False

    def __init__(
        self,
        message: str,
        nb_steps_total: int,
        nb_steps_done: int = 0,
        start_time: Optional[datetime.datetime] = None,
        time_between_reporting_s: int = 60,
        calculate_eta_since_lastreporting: bool = True,
    ):
        self.message = message
        if start_time is None:
            self.start_time = datetime.datetime.now()
        else:
            self.start_time = start_time
        self.start_time_lastreporting = self.start_time
        self.nb_steps_total = nb_steps_total
        self.nb_steps_done = nb_steps_done
        self.nb_steps_done_lastreporting = nb_steps_done
        self.time_between_reporting_s = time_between_reporting_s
        self.calculate_eta_since_lastreporting = calculate_eta_since_lastreporting

    def step(self, message: Optional[str] = None, nb_steps: int = 1):

        # Increase done counter
        self.nb_steps_done += nb_steps

        # Calculate time since last reporting
        time_now = datetime.datetime.now()
        time_passed_lastprogress_s = (
            time_now - self.start_time_lastreporting
        ).total_seconds()

        # Calculate the time_passed and nb_done we want to use to calculate ETA
        if self.calculate_eta_since_lastreporting is True:
            time_passed_for_eta_s = (
                time_now - self.start_time_lastreporting
            ).total_seconds()
            nb_steps_done_eta = self.nb_steps_done - self.nb_steps_done_lastreporting
        else:
            time_passed_for_eta_s = (time_now - self.start_time).total_seconds()
            nb_steps_done_eta = self.nb_steps_done

        # Print progress on first step or if time between reporting has passed
        if (
            time_passed_lastprogress_s >= self.time_between_reporting_s
            or self.first_reporting_done is False
        ):
            # Evade divisions by zero
            if time_passed_for_eta_s == 0 or nb_steps_done_eta == 0:
                return

            nb_per_hour = (nb_steps_done_eta / time_passed_for_eta_s) * 3600
            hours_to_go = (int)(
                (self.nb_steps_total - self.nb_steps_done) / nb_per_hour
            )
            min_to_go = (int)(
                (((self.nb_steps_total - self.nb_steps_done) / nb_per_hour) % 1) * 60
            )
            if message is not None:
                progress_message = (
                    f"{message}, {self.nb_steps_done}/{self.nb_steps_total}, "
                    f"{hours_to_go:3d}:{min_to_go:2d} left at {nb_per_hour:0.0f}/h"
                )
            else:
                progress_message = (
                    f"{self.message}, {self.nb_steps_done}/{self.nb_steps_total}, "
                    f"{hours_to_go:3d}:{min_to_go:2d} left at {nb_per_hour:0.0f}/h"
                )
            logger.info(progress_message)

            self.start_time_lastreporting = time_now
            self.nb_steps_done_lastreporting = self.nb_steps_done
            if self.first_reporting_done is False:
                self.first_reporting_done = True


# If the script is ran directly...
if __name__ == "__main__":
    raise Exception("Not implemented")



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("version.txt", mode="r") as file:
    version = file.readline()

setuptools.setup(
    name="orthoseg",
    version=version,
    author="Pieter Roggemans",
    author_email="pieter.roggemans@gmail.com",
    description="Package to make it easier to segment orthophotos.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/orthoseg/orthoseg",
    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=[
        "fiona",
        "geofileops>0.5,<0.7",
        "geopandas",
        "owslib",
        "pillow",
        "pycron",
        "rasterio",
        "segmentation-models>=1.0,<1.1",
        "simplification",
        "tensorflow>=2.5,<2.9",
    ],
    entry_points="""
            [console_scripts]
            orthoseg_load_images=orthoseg.load_images:main
            orthoseg_train=orthoseg.train:main
            orthoseg_predict=orthoseg.predict:main
            orthoseg_postprocess=orthoseg.postprocess:main
            osscriptrunner=orthoseg.scriptrunner:main
            orthoseg_load_sampleprojects=orthoseg.load_sampleprojects:main
            """,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8",
)



from pathlib import Path
from typing import Union
import requests


def download_file(
    id: str, destination: Path, ssl_verify: Union[bool, str, None] = None
):
    """
    Downloads the file to the destination filepath.

    Args
        id (str): id of the file to download.
        destination (Path): file path to download the file to.
        ssl_verify (bool or str, optional): True or None to use the default
            certificate bundle as installed on your system. False disables
            certificate validation (NOT recommended!). If a path to a
            certificate bundle file (.pem) is passed, this will be used.
            In corporate networks using a proxy server this is often needed
            to evade CERTIFICATE_VERIFY_FAILED errors. Defaults to None.
    """

    URL = "https://docs.google.com/uc?export=download"

    if ssl_verify is not None:
        # If it is a string, make sure it isn't actually a bool
        if isinstance(ssl_verify, str):
            if ssl_verify.lower() == "true":
                ssl_verify = True
            elif ssl_verify.lower() == "false":
                ssl_verify = False

        if isinstance(ssl_verify, bool) and not ssl_verify:
            print("SSL VERIFICATION IS TURNED OFF!!!")

    session = requests.Session()

    response = session.get(URL, params={"id": id}, stream=True, verify=ssl_verify)
    token = _get_confirm_token(response)

    if token:
        params = {"id": id, "confirm": token}
        response = session.get(URL, params=params, stream=True)

    destination.parent.mkdir(parents=True, exist_ok=True)
    _save_response_content(response, destination)


def _get_confirm_token(response):
    for key, value in response.cookies.items():
        if key.startswith("download_warning"):
            return value

    return None


def _save_response_content(response, destination: Path):
    CHUNK_SIZE = 32768

    with open(destination, "wb") as f:
        for chunk in response.iter_content(CHUNK_SIZE):
            if chunk:  # filter out keep-alive new chunks
                f.write(chunk)



# -*- coding: utf-8 -*-
"""
Run the scripts in a directory.
"""

import argparse
import configparser
from pathlib import Path
import subprocess
import sys
import time

# orthoseg is higher in dir hierarchy, add root to sys.path
sys.path.insert(0, str(Path(__file__).resolve().parent.parent))
from orthoseg.util import log_util
from orthoseg.util import config_util

runner_config = None


def main():

    # Interprete arguments
    parser = argparse.ArgumentParser(add_help=False)

    # Optional arguments
    optional = parser.add_argument_group("Optional arguments")
    optional.add_argument(
        "-d", "--script_dir", help="Directory containing the scripts to run."
    )
    optional.add_argument(
        "-w",
        "--watch",
        action="store_true",
        default=False,
        help="Watch the directory forever for files getting in it.",
    )
    help = "Path to a config file with parameters that need to overrule the defaults."
    optional.add_argument(
        "-c",
        "--config",
        help=help,
    )

    # Add back help
    optional.add_argument(
        "-h",
        "--help",
        action="help",
        default=argparse.SUPPRESS,
        help="Show this help message and exit",
    )
    args = parser.parse_args()

    # Init stuff
    script_dir = Path(args.script_dir)
    if not script_dir.exists():
        raise Exception(f"script dir {script_dir} does not exist")

    # Load the scriptrunner config
    conf = load_scriptrunner_config(args.config, script_dir)

    # Init logging
    log_util.clean_log_dir(
        log_dir=conf["dirs"].getpath("log_dir"),
        nb_logfiles_tokeep=conf["logging"].getint("nb_logfiles_tokeep"),
    )
    logger = log_util.init_logging_dictConfig(
        logconfig_dict=conf["logging"].getdict("logconfig"),
        log_basedir=conf["dirs"].getpath("script_dir"),
        loggername=__name__,
    )

    # Init working dirs
    done_dir = conf["dirs"].getpath("done_dir")
    error_dir = conf["dirs"].getpath("error_dir")

    # Loop over scripts to be ran
    wait_message_printed = False
    while True:

        # List the scripts in the dir
        script_paths = []
        script_patterns = conf["general"].getlist("script_patterns")
        for script_pattern in script_patterns:
            script_paths.extend(list(script_dir.glob(script_pattern)))

        # If no scripts found, sleep or stop...
        if len(script_paths) == 0:
            if args.watch is False:
                logger.info(f"No scripts found (anymore) in {script_dir}, so stop")
                break
            else:
                if wait_message_printed is False:
                    logger.info(
                        f"No scripts to run in {script_dir}, so watch script dir..."
                    )
                    wait_message_printed = True
                time.sleep(10)
                continue

        # Get next script alphabetically
        script_path = sorted(script_paths)[0]

        try:
            # Run the script and print output in realtime
            wait_message_printed = False
            logger.info(f"Run script {script_path}")
            cmd = [script_path]

            process = subprocess.Popen(
                cmd,
                shell=True,
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                text=True,
                encoding="utf-8",
            )
            # ,creationflags=subprocess.CREATE_NO_WINDOW)

            still_running = True
            while still_running:
                if process.stdout is not None:
                    output = process.stdout.readline()
                    if output == "" and process.poll() is not None:
                        still_running = False
                    if output:
                        logger.info(output.strip())

            # If error code != 0, an error occured
            rc = process.poll()
            if rc != 0:
                # Script gave an error, so move to error dir
                logger.error(f"Script {script_path} gave error return code: {rc}")
                error_dir.mkdir(parents=True, exist_ok=True)
                error_path = error_dir / script_path.name
                if error_path.exists():
                    error_path.unlink()
                script_path.rename(target=error_path)
            else:
                # Move the script to the done dir
                done_dir.mkdir(parents=True, exist_ok=True)
                done_path = done_dir / script_path.name
                if done_path.exists():
                    done_path.unlink()
                script_path.rename(target=done_path)

        except Exception:
            logger.exception(f"Error running script {script_path}")

            # If the script still exists, move it to error dir
            if script_path.exists():
                error_dir.mkdir(parents=True, exist_ok=True)
                error_path = error_dir / script_path.name
                if error_path.exists():
                    error_path.unlink()
                script_path.rename(target=error_path)


def load_scriptrunner_config(
    config_path: str, script_dir: Path
) -> configparser.ConfigParser:

    # Load defaults first
    scriptrunner_py_dir = Path(__file__).resolve().parent
    config_paths = [scriptrunner_py_dir / "scriptrunner_defaults.ini"]

    # If a config path is specified, this config should overrule the defaults
    if config_path is not None:
        config_paths.append(Path(config_path))

    # Load!
    scriptrunner_config = config_util.read_config_ext(config_paths)

    # If a script_dir is specified, it overrides the config file(s)
    if script_dir is not None:
        scriptrunner_config["dirs"]["script_dir"] = script_dir.as_posix()

    return scriptrunner_config


if __name__ == "__main__":
    main()
